"""Constants used across the project."""

from rq_geo_toolkit.constants import (
    GEOMETRY_COLUMN,
    PARQUET_COMPRESSION,
    PARQUET_COMPRESSION_LEVEL,
    PARQUET_ROW_GROUP_SIZE,
)

WGS84_CRS = "EPSG:4326"

FEATURES_INDEX = "feature_id"

METADATA_TAGS_TO_IGNORE = [
    "area",
    "created_by",
    "converted_by",
    "source",
    "time",
    "ele",
    "note",
    "todo",
    "fixme",
    "FIXME",
    "openGeoDB:",
]

__all__ = [
    "FEATURES_INDEX",
    "GEOMETRY_COLUMN",
    "PARQUET_COMPRESSION",
    "PARQUET_COMPRESSION_LEVEL",
    "PARQUET_ROW_GROUP_SIZE",
    "WGS84_CRS",
]



"""Global conftest for QuackOSM tests."""

import os
import shutil
from pathlib import Path

import pytest
from pytest import Item


def pytest_runtest_setup(item: Item) -> None:
    """Setup python encoding before `pytest_runtest_call(item)`."""
    os.environ["PYTHONIOENCODING"] = "utf-8"


@pytest.fixture(autouse=True, scope="session")  # type: ignore
def copy_geocode_cache() -> None:
    """Load cached geocoding results."""
    existing_cache_directory = Path(__file__).parent / "test_files" / "geocoding_cache"
    geocoding_cache_directory = Path("cache")
    geocoding_cache_directory.mkdir(exist_ok=True)
    for file_path in existing_cache_directory.glob("*.json"):
        destination_path = geocoding_cache_directory / file_path.name
        shutil.copy(file_path, destination_path)



"""Tests for CLI."""

import uuid
from pathlib import Path
from typing import Optional

import pytest
from parametrization import Parametrization as P
from pytest_mock import MockerFixture
from typer.testing import CliRunner

from quackosm import __app_name__, __version__, cli
from quackosm.osm_extracts.extract import OsmExtractSource
from tests.base.conftest import geometry_boundary_file_path, geometry_geojson, geometry_wkt

runner = CliRunner()


def monaco_pbf_file_path() -> str:
    """Monaco PBF file path."""
    return str(Path(__file__).parent.parent / "test_files" / "monaco.osm.pbf")


@pytest.fixture()  # type: ignore
def monaco_pbf_file_path_fixture() -> str:
    """Monaco PBF file path fixture."""
    return monaco_pbf_file_path()


def osm_tags_filter_file_path() -> str:
    """OSM tags filter file path."""
    return str(Path(__file__).parent.parent / "test_files" / "osm_tags_filter.json")


def osm_way_config_file_path() -> str:
    """OSM way features config file path."""
    return str(Path(__file__).parent.parent.parent / "quackosm" / "osm_way_polygon_features.json")


def random_str() -> str:
    """Return random string."""
    return str(uuid.uuid4())


def test_version() -> None:
    """Test if version is properly returned."""
    result = runner.invoke(cli.app, ["--version"])

    assert result.exit_code == 0
    assert f"{__app_name__} {__version__}\n" in result.stdout


def test_pbf_file_or_geometry_filter_is_required() -> None:
    """Test if cannot run without pbf file and without geometry filter."""
    result = runner.invoke(
        cli.app,
    )

    assert result.exit_code == 2
    assert "Missing argument 'PBF file path'." in result.stdout


def test_basic_run(monaco_pbf_file_path_fixture: str) -> None:
    """Test if runs properly without options."""
    result = runner.invoke(cli.app, [monaco_pbf_file_path_fixture])

    assert result.exit_code == 0
    assert str(Path("files/monaco_nofilter_noclip_compact_sorted.parquet")) in result.stdout


def test_silent_mode(monaco_pbf_file_path_fixture: str) -> None:
    """Test if runs properly without reporting status."""
    result = runner.invoke(cli.app, [monaco_pbf_file_path_fixture, "--ignore-cache", "--silent"])

    assert result.exit_code == 0
    assert str(Path("files/monaco_nofilter_noclip_compact_sorted.parquet")) == result.stdout.strip()


def test_transient_mode(monaco_pbf_file_path_fixture: str) -> None:
    """Test if runs properly without reporting status."""
    result = runner.invoke(cli.app, [monaco_pbf_file_path_fixture, "--ignore-cache", "--transient"])
    output_lines = result.stdout.strip().split("\n")
    assert result.exit_code == 0
    assert len(result.stdout.strip().split("\n")) == 2
    assert "Finished operation in" in output_lines[0]
    assert str(Path("files/monaco_nofilter_noclip_compact_sorted.parquet")) == output_lines[1]


@P.parameters("args", "expected_result")  # type: ignore
@P.case(
    "Explode",
    ["--explode-tags"],
    "files/monaco_nofilter_noclip_exploded_sorted.parquet",
)  # type: ignore
@P.case("Explode short", ["--explode"], "files/monaco_nofilter_noclip_exploded_sorted.parquet")  # type: ignore
@P.case("Compact", ["--compact-tags"], "files/monaco_nofilter_noclip_compact_sorted.parquet")  # type: ignore
@P.case("Compact short", ["--compact"], "files/monaco_nofilter_noclip_compact_sorted.parquet")  # type: ignore
@P.case(
    "Working directory",
    ["--working-directory", "files/workdir"],
    "files/workdir/monaco_nofilter_noclip_compact_sorted.parquet",
)  # type: ignore
@P.case("Ignore cache", ["--ignore-cache"], "files/monaco_nofilter_noclip_compact_sorted.parquet")  # type: ignore
@P.case("Ignore cache short", ["--no-cache"], "files/monaco_nofilter_noclip_compact_sorted.parquet")  # type: ignore
@P.case("Output", ["--output", "files/monaco_output.parquet"], "files/monaco_output.parquet")  # type: ignore
@P.case("Output short", ["-o", "files/monaco_output.parquet"], "files/monaco_output.parquet")  # type: ignore
@P.case(
    "DuckDB explicit export",
    ["--duckdb"],
    "files/monaco_nofilter_noclip_compact_sorted.duckdb",
)  # type: ignore
@P.case(
    "DuckDB explicit export with table name",
    ["--duckdb", "--duckdb-table-name", "test"],
    "files/monaco_nofilter_noclip_compact_sorted.duckdb",
)  # type: ignore
@P.case("Silent", ["--silent"], "files/monaco_nofilter_noclip_compact_sorted.parquet")  # type: ignore
@P.case("Transient", ["--transient"], "files/monaco_nofilter_noclip_compact_sorted.parquet")  # type: ignore
@P.case("Explicit sort", ["--sort"], "files/monaco_nofilter_noclip_compact_sorted.parquet")  # type: ignore
@P.case("No sort", ["--no-sort"], "files/monaco_nofilter_noclip_compact.parquet")  # type: ignore
@P.case(
    "Keep metadata tags",
    ["--keep-metadata-tags"],
    "files/monaco_d38c1671_noclip_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Skip metadata tags",
    ["--ignore-metadata-tags"],
    "files/monaco_nofilter_noclip_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Output with working directory",
    ["--working-directory", "files/workdir", "-o", "files/monaco_output.parquet"],
    "files/monaco_output.parquet",
)  # type: ignore
@P.case(
    "OSM tags filter",
    [
        "--osm-tags-filter",
        '{"building": true, "highway": ["primary", "secondary"], "amenity": "bench"}',
    ],
    "files/monaco_a9dd1c3c_noclip_exploded_sorted.parquet",
)  # type: ignore
@P.case(
    "OSM tags filter compact",
    [
        "--osm-tags-filter",
        '{"building": true, "highway": ["primary", "secondary"], "amenity": "bench"}',
        "--compact",
    ],
    "files/monaco_a9dd1c3c_noclip_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "OSM tags filter file",
    [
        "--osm-tags-filter-file",
        osm_tags_filter_file_path(),
    ],
    "files/monaco_a9dd1c3c_noclip_exploded_sorted.parquet",
)  # type: ignore
@P.case(
    "OSM tags filter file compact",
    [
        "--osm-tags-filter-file",
        osm_tags_filter_file_path(),
        "--compact",
    ],
    "files/monaco_a9dd1c3c_noclip_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "OSM tags filter grouped",
    [
        "--osm-tags-filter",
        '{"group": {"building": true, "highway": ["primary", "secondary"], "amenity": "bench"} }',
    ],
    "files/monaco_654daac5_noclip_exploded_sorted.parquet",
)  # type: ignore
@P.case(
    "OSM tags filter grouped compact",
    [
        "--osm-tags-filter",
        '{"group": {"building": true, "highway": ["primary", "secondary"], "amenity": "bench"} }',
        "--compact",
    ],
    "files/monaco_654daac5_noclip_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Geometry WKT filter",
    ["--geom-filter-wkt", geometry_wkt()],
    "files/monaco_nofilter_09c3fc04_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Geometry GeoJSON filter",
    ["--geom-filter-geojson", geometry_geojson()],
    "files/monaco_nofilter_82c0fdfa_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Geometry file filter",
    ["--geom-filter-file", geometry_boundary_file_path()],
    "files/monaco_nofilter_6a869bcf_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Geometry geocode filter",
    ["--geom-filter-geocode", "Monaco-Ville, Monaco"],
    "files/monaco_nofilter_e7f0b78a_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Geometry Geohash filter",
    ["--geom-filter-index-geohash", "spv2bc"],
    "files/monaco_nofilter_c08889e8_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Geometry Geohash filter multiple",
    ["--geom-filter-index-geohash", "spv2bc,spv2bfr"],
    "files/monaco_nofilter_1bd33e0a_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Geometry H3 filter",
    ["--geom-filter-index-h3", "8a3969a40ac7fff"],
    "files/monaco_nofilter_01d82ad1_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Geometry H3 filter multiple",
    ["--geom-filter-index-h3", "8a3969a40ac7fff,893969a4037ffff"],
    "files/monaco_nofilter_c3252c58_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Geometry S2 filter",
    ["--geom-filter-index-s2", "12cdc28bc"],
    "files/monaco_nofilter_5c3d61eb_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Geometry S2 filter multiple",
    ["--geom-filter-index-s2", "12cdc28bc,12cdc28f"],
    "files/monaco_nofilter_cda5d65e_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Filter OSM features IDs",
    ["--filter-osm-ids", "way/94399646,node/3617982224,relation/36990"],
    "files/monaco_nofilter_noclip_compact_c740a159_sorted.parquet",
)  # type: ignore
@P.case(
    "Custom SQL filter",
    ["--custom-sql-filter", "cardinality(tags) = 5"],
    "files/monaco_138f715c_noclip_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Custom SQL filter",
    ["--custom-sql-filter", "tags['highway'][1] = 'primary'"],
    "files/monaco_e1a5cc12_noclip_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Keep all tags",
    [
        "--keep-all-tags",
    ],
    "files/monaco_nofilter_noclip_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "OSM tags filter with keep all tags",
    [
        "--keep-all-tags",
        "--osm-tags-filter",
        '{"building": true, "highway": ["primary", "secondary"], "amenity": "bench"}',
    ],
    "files/monaco_a9dd1c3c_alltags_noclip_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "OSM tags filter with keep all tags compact",
    [
        "--keep-all-tags",
        "--osm-tags-filter",
        '{"building": true, "highway": ["primary", "secondary"], "amenity": "bench"}',
        "--compact",
    ],
    "files/monaco_a9dd1c3c_alltags_noclip_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "OSM tags filter with keep all tags exploded",
    [
        "--keep-all-tags",
        "--osm-tags-filter",
        '{"building": true, "highway": ["primary", "secondary"], "amenity": "bench"}',
        "--explode",
    ],
    "files/monaco_a9dd1c3c_alltags_noclip_exploded_sorted.parquet",
)  # type: ignore
@P.case(
    "OSM way polygon config",
    ["--osm-way-polygon-config", osm_way_config_file_path()],
    "files/monaco_nofilter_noclip_compact_sorted.parquet",
)  # type: ignore
@P.case("WKT", ["--wkt-result"], "files/monaco_nofilter_noclip_compact_wkt.parquet")  # type: ignore
@P.case("WKT short", ["--wkt"], "files/monaco_nofilter_noclip_compact_wkt.parquet")  # type: ignore
def test_proper_args_with_pbf(
    monaco_pbf_file_path_fixture: str, args: list[str], expected_result: str
) -> None:
    """Test if runs properly with options."""
    result = runner.invoke(cli.app, [monaco_pbf_file_path_fixture, *args])
    print(result.stdout)

    assert result.exit_code == 0
    assert str(Path(expected_result)) in result.stdout


@P.parameters("args", "expected_result")  # type: ignore
@P.case(
    "Geometry BBOX filter",
    ["--geom-filter-bbox", "7.416486,43.731086,7.421931,43.733707"],
    "files/b9115f99_nofilter_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Geometry WKT filter",
    ["--geom-filter-wkt", geometry_wkt()],
    "files/09c3fc04_nofilter_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Geometry GeoJSON filter",
    ["--geom-filter-geojson", geometry_geojson()],
    "files/82c0fdfa_nofilter_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Geometry geocode filter",
    ["--geom-filter-geocode", "Monaco-Ville, Monaco"],
    "files/e7f0b78a_nofilter_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Geometry Geohash filter",
    ["--geom-filter-index-geohash", "spv2bc"],
    "files/c08889e8_nofilter_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Geometry Geohash filter multiple",
    ["--geom-filter-index-geohash", "spv2bc,spv2bfr"],
    "files/1bd33e0a_nofilter_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Geometry H3 filter",
    ["--geom-filter-index-h3", "8a3969a40ac7fff"],
    "files/01d82ad1_nofilter_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Geometry H3 filter multiple",
    ["--geom-filter-index-h3", "8a3969a40ac7fff,893969a4037ffff"],
    "files/c3252c58_nofilter_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Geometry S2 filter",
    ["--geom-filter-index-s2", "12cdc28bc"],
    "files/5c3d61eb_nofilter_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Geometry S2 filter multiple",
    ["--geom-filter-index-s2", "12cdc28bc,12cdc28f"],
    "files/cda5d65e_nofilter_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Geometry file filter with different OSM source",
    ["--geom-filter-file", geometry_boundary_file_path(), "--osm-extract-source", "OSMfr"],
    "files/6a869bcf_nofilter_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Explode",
    ["--geom-filter-file", geometry_boundary_file_path(), "--explode-tags"],
    "files/6a869bcf_nofilter_exploded_sorted.parquet",
)  # type: ignore
@P.case(
    "Compact",
    ["--geom-filter-file", geometry_boundary_file_path(), "--compact-tags"],
    "files/6a869bcf_nofilter_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Ignore cache",
    ["--geom-filter-file", geometry_boundary_file_path(), "--ignore-cache"],
    "files/6a869bcf_nofilter_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Explicit sort",
    ["--geom-filter-file", geometry_boundary_file_path(), "--sort"],
    "files/6a869bcf_nofilter_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "No sort",
    ["--geom-filter-file", geometry_boundary_file_path(), "--no-sort"],
    "files/6a869bcf_nofilter_compact.parquet",
)  # type: ignore
@P.case(
    "Keep metadata tags",
    ["--geom-filter-file", geometry_boundary_file_path(), "--keep-metadata-tags"],
    "files/6a869bcf_d38c1671_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Skip metadata tags",
    ["--geom-filter-file", geometry_boundary_file_path(), "--ignore-metadata-tags"],
    "files/6a869bcf_nofilter_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Working directory",
    [
        "--geom-filter-file",
        geometry_boundary_file_path(),
        "--working-directory",
        "files/workdir",
        "--ignore-cache",
    ],
    "files/workdir/6a869bcf_nofilter_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Output",
    [
        "--geom-filter-file",
        geometry_boundary_file_path(),
        "--output",
        "files/monaco_output.parquet",
    ],
    "files/monaco_output.parquet",
)  # type: ignore
@P.case(
    "Output with working directory",
    [
        "--geom-filter-file",
        geometry_boundary_file_path(),
        "--working-directory",
        "files/workdir",
        "-o",
        "files/monaco_output.parquet",
    ],
    "files/monaco_output.parquet",
)  # type: ignore
@P.case(
    "OSM tags filter",
    [
        "--geom-filter-file",
        geometry_boundary_file_path(),
        "--osm-tags-filter",
        '{"building": true, "highway": ["primary", "secondary"], "amenity": "bench"}',
    ],
    "files/6a869bcf_a9dd1c3c_exploded_sorted.parquet",
)  # type: ignore
@P.case(
    "OSM tags filter compact",
    [
        "--geom-filter-file",
        geometry_boundary_file_path(),
        "--osm-tags-filter",
        '{"building": true, "highway": ["primary", "secondary"], "amenity": "bench"}',
        "--compact",
    ],
    "files/6a869bcf_a9dd1c3c_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "OSM tags filter grouped",
    [
        "--geom-filter-file",
        geometry_boundary_file_path(),
        "--osm-tags-filter",
        '{"group": {"building": true, "highway": ["primary", "secondary"], "amenity": "bench"} }',
    ],
    "files/6a869bcf_654daac5_exploded_sorted.parquet",
)  # type: ignore
@P.case(
    "OSM tags filter grouped compact",
    [
        "--geom-filter-file",
        geometry_boundary_file_path(),
        "--osm-tags-filter",
        '{"group": {"building": true, "highway": ["primary", "secondary"], "amenity": "bench"} }',
        "--compact",
    ],
    "files/6a869bcf_654daac5_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Filter OSM features IDs",
    [
        "--geom-filter-file",
        geometry_boundary_file_path(),
        "--filter-osm-ids",
        "way/94399646,node/3617982224,relation/36990",
    ],
    "files/6a869bcf_nofilter_compact_c740a159_sorted.parquet",
)  # type: ignore
@P.case(
    "Keep all tags",
    [
        "--geom-filter-file",
        geometry_boundary_file_path(),
        "--keep-all-tags",
    ],
    "files/6a869bcf_nofilter_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "OSM tags filter with keep all tags",
    [
        "--geom-filter-file",
        geometry_boundary_file_path(),
        "--keep-all-tags",
        "--osm-tags-filter",
        '{"building": true, "highway": ["primary", "secondary"], "amenity": "bench"}',
    ],
    "files/6a869bcf_a9dd1c3c_alltags_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "OSM tags filter with keep all tags compact",
    [
        "--geom-filter-file",
        geometry_boundary_file_path(),
        "--keep-all-tags",
        "--osm-tags-filter",
        '{"building": true, "highway": ["primary", "secondary"], "amenity": "bench"}',
        "--compact",
    ],
    "files/6a869bcf_a9dd1c3c_alltags_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "OSM tags filter with keep all tags exploded",
    [
        "--geom-filter-file",
        geometry_boundary_file_path(),
        "--keep-all-tags",
        "--osm-tags-filter",
        '{"building": true, "highway": ["primary", "secondary"], "amenity": "bench"}',
        "--explode",
    ],
    "files/6a869bcf_a9dd1c3c_alltags_exploded_sorted.parquet",
)  # type: ignore
@P.case(
    "OSM way polygon config",
    [
        "--geom-filter-file",
        geometry_boundary_file_path(),
        "--osm-way-polygon-config",
        osm_way_config_file_path(),
    ],
    "files/6a869bcf_nofilter_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Allow not covered geometry",
    [
        "--geom-filter-wkt",
        (
            "POLYGON ((-43.064 29.673, -43.064 29.644, -43.017 29.644,"
            " -43.017 29.673, -43.064 29.673))"
        ),
        "--allow-uncovered-geometry",
        "--ignore-cache",
    ],
    "files/fa44926c_nofilter_compact_sorted.parquet",
)  # type: ignore
def test_proper_args_with_geometry_filter(args: list[str], expected_result: str) -> None:
    """Test if runs properly with options."""
    result = runner.invoke(cli.app, [*args, "--osm-extract-source", "any"])
    print(result.stdout)

    assert result.exit_code == 0
    assert str(Path(expected_result)) in result.stdout


@P.parameters("args", "expected_result")  # type: ignore
@P.case(
    "Full name",
    [
        "--osm-extract-query",
        "geofabrik_europe_monaco",
    ],
    "files/geofabrik_europe_monaco_nofilter_noclip_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Find in geofabrik",
    [
        "--osm-extract-query",
        "Monaco",
        "--osm-extract-source",
        "Geofabrik",
    ],
    "files/geofabrik_europe_monaco_nofilter_noclip_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Explode",
    ["--osm-extract-query", "geofabrik_europe_monaco", "--explode-tags"],
    "files/geofabrik_europe_monaco_nofilter_noclip_exploded_sorted.parquet",
)  # type: ignore
@P.case(
    "Compact",
    [
        "--osm-extract-query",
        "geofabrik_europe_monaco",
        "--compact-tags",
    ],
    "files/geofabrik_europe_monaco_nofilter_noclip_compact_sorted.parquet",
)  # type: ignore
@P.case(
    "Output",
    [
        "--osm-extract-query",
        "geofabrik_europe_monaco",
        "--ignore-cache",
        "--output",
        "files/gfbrk_monaco_output.parquet",
    ],
    "files/gfbrk_monaco_output.parquet",
)  # type: ignore
@P.case(
    "OSM tags filter and geometry filter",
    [
        "--osm-extract-query",
        "geofabrik_europe_monaco",
        "--geom-filter-file",
        geometry_boundary_file_path(),
        "--osm-tags-filter",
        '{"building": true, "highway": ["primary", "secondary"], "amenity": "bench"}',
    ],
    "files/geofabrik_europe_monaco_a9dd1c3c_6a869bcf_exploded_sorted.parquet",
)  # type: ignore
def test_proper_args_with_osm_extract(args: list[str], expected_result: str) -> None:
    """Test if runs properly with options."""
    result = runner.invoke(cli.app, [*args])
    print(result.stdout)

    assert result.exit_code == 0
    assert str(Path(expected_result)) in result.stdout


def test_proper_args_with_pbf_url() -> None:
    """Test if runs properly with an url path."""
    result = runner.invoke(cli.app, ["https://download.geofabrik.de/europe/monaco-latest.osm.pbf"])
    print(result.stdout)

    assert result.exit_code == 0
    assert str(Path("files/monaco-latest_nofilter_noclip_compact_sorted.parquet")) in result.stdout


@P.parameters("args")  # type: ignore
@P.case(
    "OSM tags filter malfunctioned JSON",
    [
        monaco_pbf_file_path(),
        "--osm-tags-filter",
        '{"building": True, "highway": ["primary", "secondary"], "amenity": "bench"}',
    ],
)  # type: ignore
@P.case(
    "OSM tags filter malfunctioned JSON",
    [
        monaco_pbf_file_path(),
        "--osm-tags-filter",
        '{"building": true, highway": ["primary", "secondary"], "amenity": "bench"}',
    ],
)  # type: ignore
@P.case(
    "OSM tags filter malfunctioned JSON",
    [
        monaco_pbf_file_path(),
        "--osm-tags-filter",
        '{"building": true, "highway": ["primary", "secondary"], "amenity": "bench"',
    ],
)  # type: ignore
@P.case(
    "OSM tags filter wrong type",
    [
        monaco_pbf_file_path(),
        "--osm-tags-filter",
        (
            '{"super_group": {"group": {"building": true, "highway": ["primary", "secondary"],'
            ' "amenity": "bench"} } }'
        ),
    ],
)  # type: ignore
@P.case(
    "OSM tags filter wrong type",
    [
        monaco_pbf_file_path(),
        "--osm-tags-filter",
        '{"group": [{"building": true, "highway": ["primary", "secondary"], "amenity": "bench"}] }',
    ],
)  # type: ignore
@P.case(
    "OSM tags two filters",
    [
        monaco_pbf_file_path(),
        "--osm-tags-filter",
        '{"building": true, "highway": ["primary", "secondary"], "amenity": "bench"}',
        "--osm-tags-filter-file",
        osm_tags_filter_file_path(),
    ],
)  # type: ignore
@P.case(
    "OSM tags nonexistent file filter",
    [monaco_pbf_file_path(), "--osm-tags-filter-file", "nonexistent_json_file.json"],
)  # type: ignore
@P.case("Geometry WKT filter with GeoJSON", ["--geom-filter-wkt", geometry_geojson()])  # type: ignore
@P.case("Geometry GeoJSON filter with WKT", ["--geom-filter-geojson", geometry_wkt()])  # type: ignore
@P.case("Geometry Geohash filter with random string", ["--geom-filter-index-geohash", random_str()])  # type: ignore
@P.case("Geometry H3 filter with Geohash", ["--geom-filter-index-h3", "spv2bc"])  # type: ignore
@P.case("Geometry H3 filter with S2", ["--geom-filter-index-h3", "12cdc28bc"])  # type: ignore
@P.case("Geometry H3 filter with random string", ["--geom-filter-index-h3", random_str()])  # type: ignore
@P.case("Geometry S2 filter with random string", ["--geom-filter-index-s2", random_str()])  # type: ignore
@P.case("Geometry BBOX filter with wrong values", ["--geom-filter-bbox", random_str()])  # type: ignore
@P.case("Geometry BBOX filter with spaces", ["--geom-filter-bbox", "10,", "-5,", "6,", "17"])  # type: ignore
@P.case(
    "Geometry two filters",
    ["--geom-filter-geojson", geometry_geojson(), "--geom-filter-wkt", geometry_wkt()],
)  # type: ignore
@P.case(
    "Geometry nonexistent file filter",
    ["--geom-filter-file", "nonexistent_geojson_file.geojson"],
)  # type: ignore
@P.case(
    "Geometry wrong file filter",
    ["--geom-filter-file", osm_tags_filter_file_path()],
)  # type: ignore
@P.case("Filter OSM", [monaco_pbf_file_path(), "--filter-osm-ids", "124"])  # type: ignore
@P.case("Filter OSM", [monaco_pbf_file_path(), "--filter-osm-ids", "w/124"])  # type: ignore
@P.case("Filter OSM", [monaco_pbf_file_path(), "--filter-osm-ids", "w124"])  # type: ignore
@P.case("Filter OSM", [monaco_pbf_file_path(), "--filter-osm-ids", "way124"])  # type: ignore
@P.case("Filter OSM", [monaco_pbf_file_path(), "--filter-osm-ids", "n/124"])  # type: ignore
@P.case("Filter OSM", [monaco_pbf_file_path(), "--filter-osm-ids", "n124"])  # type: ignore
@P.case("Filter OSM", [monaco_pbf_file_path(), "--filter-osm-ids", "node124"])  # type: ignore
@P.case("Filter OSM", [monaco_pbf_file_path(), "--filter-osm-ids", "r/124"])  # type: ignore
@P.case("Filter OSM", [monaco_pbf_file_path(), "--filter-osm-ids", "r124"])  # type: ignore
@P.case("Filter OSM", [monaco_pbf_file_path(), "--filter-osm-ids", "relation124"])  # type: ignore
@P.case("Filter OSM", [monaco_pbf_file_path(), "--filter-osm-ids", "node/124;way/124;relation/124"])  # type: ignore
@P.case("Filter OSM", [monaco_pbf_file_path(), "--filter-osm-ids", "node/124|way/124|relation/124"])  # type: ignore
@P.case("Filter OSM", [monaco_pbf_file_path(), "--filter-osm-ids", "node/124 way/124 relation/124"])  # type: ignore
@P.case("Custom SQL filter", [monaco_pbf_file_path(), "--custom-sql-filter", random_str()])  # type: ignore
@P.case(
    "Custom SQL filter", [monaco_pbf_file_path(), "--custom-sql-filter", "cardinality(map) = 5"]
)  # type: ignore
@P.case(
    "Custom SQL filter",
    [monaco_pbf_file_path(), "--custom-sql-filter", "map['highway'][1] = 'primary'"],
)  # type: ignore
@P.case(
    "OSM way polygon config",
    [monaco_pbf_file_path(), "--osm-way-polygon-config", "nonexistent_json_file.json"],
)  # type: ignore
@P.case("Silent and transient", [monaco_pbf_file_path(), "--silent", "--transient"])  # type: ignore
@P.case("OSM extracts with multiple matches", ["--osm-extract-query", "Monaco"])  # type: ignore
@P.case("OSM extracts with zero matches - some close matches", ["--osm-extract-query", "Prlnd"])  # type: ignore
@P.case(
    "OSM extracts with zero matches - without close matches",
    [
        "--osm-extract-query",
        "nonexistent_extract",
    ],
)  # type: ignore
@P.case(
    "OSM extracts with zero matches and duckdb export",
    [
        "--duckdb",
        "--osm-extract-query",
        "quack_extract",
    ],
)  # type: ignore
@P.case(
    "Wrong IoU threshold value",
    [
        "--iou-threshold",
        "1.2",
    ],
)  # type: ignore
@P.case(
    "Wrong IoU threshold value",
    [
        "--iou-threshold",
        "-0.2",
    ],
)  # type: ignore
def test_wrong_args(args: list[str], capsys: pytest.CaptureFixture) -> None:
    """Test if doesn't run properly with options."""
    # Fix for the I/O error from the Click repository
    # https://github.com/pallets/click/issues/824#issuecomment-1583293065
    with capsys.disabled():
        result = runner.invoke(cli.app, args)
        assert result.exit_code != 0


@pytest.mark.parametrize(
    "osm_source",
    [None, *list(OsmExtractSource)],
)  # type: ignore
@pytest.mark.parametrize(
    "command",
    ["show-extracts", "show-osm-extracts"],
)  # type: ignore
def test_displaying_osm_extracts(
    command: str,
    osm_source: Optional[OsmExtractSource],
    mocker: MockerFixture,
    capsys: pytest.CaptureFixture,
) -> None:
    """Test if displaying OSM extracts works."""
    with capsys.disabled():
        osm_source_command = ["--osm-extract-source", osm_source.value] if osm_source else []
        result = runner.invoke(cli.app, [f"--{command}", *osm_source_command])
        output = result.stdout

        assert result.exit_code == 0
        assert len(output) > 0

        osm_sources_without_any = [src for src in OsmExtractSource if src != OsmExtractSource.any]

        if osm_source == OsmExtractSource.any or not osm_source:
            assert output.startswith("All extracts")
            assert all(src.value in output for src in osm_sources_without_any)
        else:
            assert output.startswith(osm_source.value)

        lines = output.lower().split("\n")

        assert all(
            any(src.value.lower() in line for src in osm_sources_without_any)
            for line in lines
            if len(line.strip()) > 0 and line != "all extracts"
        )



"""
BBBike OpenStreetMap extracts.

This module contains wrapper for publically available BBBike download server.
"""

from typing import Optional

import geopandas as gpd
import requests
from tqdm import tqdm

from quackosm._rich_progress import FORCE_TERMINAL
from quackosm.osm_extracts._poly_parser import parse_polygon_file
from quackosm.osm_extracts.extract import (
    OpenStreetMapExtract,
    OsmExtractSource,
    extracts_to_geodataframe,
    load_index_decorator,
)

BBBIKE_EXTRACTS_INDEX_URL = "https://download.bbbike.org/osm/bbbike"
BBBIKE_INDEX_GDF: Optional[gpd.GeoDataFrame] = None

__all__ = ["_get_bbbike_index"]


def _get_bbbike_index() -> gpd.GeoDataFrame:
    global BBBIKE_INDEX_GDF  # noqa: PLW0603

    if BBBIKE_INDEX_GDF is None:
        BBBIKE_INDEX_GDF = _load_bbbike_index()

    return BBBIKE_INDEX_GDF


@load_index_decorator(OsmExtractSource.bbbike)
def _load_bbbike_index() -> gpd.GeoDataFrame:  # pragma: no cover
    """
    Load available extracts from BBBike download service.

    Returns:
        gpd.GeoDataFrame: Extracts index with metadata.
    """
    extracts = _iterate_bbbike_index()
    gdf = extracts_to_geodataframe(extracts)

    return gdf


def _iterate_bbbike_index() -> list[OpenStreetMapExtract]:  # pragma: no cover
    """
    Iterate OpenStreetMap.fr extracts service page.

    Works recursively, by scraping whole available directory.

    Returns:
        List[OpenStreetMapExtract]: List of loaded osm.fr extracts objects.
    """
    from bs4 import BeautifulSoup

    extracts = []
    result = requests.get(
        BBBIKE_EXTRACTS_INDEX_URL,
        headers={"User-Agent": "QuackOSM Python package (https://github.com/kraina-ai/quackosm)"},
    )
    soup = BeautifulSoup(result.text, "html.parser")
    extract_names = [
        extract_href.text
        for extract_href in soup.select("tr.d > td > a")
        if extract_href.text != ".."
    ]

    bbbike_enum_value = OsmExtractSource.bbbike.value

    with tqdm(
        disable=FORCE_TERMINAL, desc=bbbike_enum_value, total=len(extract_names)
    ) as pbar:
        for extract_name in extract_names:
            pbar.set_description(f"{bbbike_enum_value}_{extract_name}")
            poly_url = f"{BBBIKE_EXTRACTS_INDEX_URL}/{extract_name}/{extract_name}.poly"
            polygon = parse_polygon_file(poly_url)
            if polygon is None:
                continue
            pbf_url = f"{BBBIKE_EXTRACTS_INDEX_URL}/{extract_name}/{extract_name}.osm.pbf"
            extracts.append(
                OpenStreetMapExtract(
                    id=f"{bbbike_enum_value}_{extract_name}",
                    name=extract_name,
                    parent=bbbike_enum_value,
                    url=pbf_url,
                    geometry=polygon,
                )
            )
            pbar.update()

    return extracts



"""Poly file parser function."""

from typing import Any, Optional

import requests
from shapely.geometry import MultiPolygon

__all__ = ["parse_polygon_file"]


def parse_polygon_file(polygon_url: str) -> Optional[MultiPolygon]:  # pragma: no cover
    """
    Parse poly file from URL to geometry.

    Args:
        polygon_url (str): URL to load a poly file.

    Returns:
        Optional[MultiPolygon]: Parsed polygon.
            Empty if request returns 404 not found.
    """
    result = requests.get(
        polygon_url,
        headers={"User-Agent": "QuackOSM Python package (https://github.com/kraina-ai/quackosm)"},
    )
    if result.status_code == 404:
        return None
    result.raise_for_status()
    poly = parse_poly(result.text.splitlines())
    return poly


def parse_poly(lines: list[str]) -> MultiPolygon:  # pragma: no cover
    """
    Parse an Osmosis polygon filter file.

    Accept a sequence of lines from a polygon file, return a shapely.geometry.MultiPolygon object.
    Based on: https://wiki.openstreetmap.org/wiki/Osmosis/Polygon_Filter_File_Python_Parsing

    http://wiki.openstreetmap.org/wiki/Osmosis/Polygon_Filter_File_Format
    """
    in_ring = False
    coords: list[Any] = []

    for index, line in enumerate(lines):
        if index == 0:
            # first line is junk.
            continue

        elif index == 1:
            # second line is the first polygon ring.
            coords.append([[], []])
            ring = coords[-1][0]
            in_ring = True

        elif in_ring and line.strip() == "END":
            # we are at the end of a ring, perhaps with more to come.
            in_ring = False

        elif in_ring:
            # we are in a ring and picking up new coordinates.
            ring.append(list(map(float, line.split())))

        elif not in_ring and line.strip() == "END":
            # we are at the end of the whole polygon.
            break

        elif not in_ring and line.startswith("!"):
            # we are at the start of a polygon part hole.
            coords[-1][1].append([])
            ring = coords[-1][1][-1]
            in_ring = True

        elif not in_ring:
            # we are at the start of a polygon part.
            coords.append([[], []])
            ring = coords[-1][0]
            in_ring = True

    return MultiPolygon(coords)



"""Wrapper over Rich progress bar."""

import json
import time
from collections.abc import Generator, Iterable
from datetime import timedelta
from types import TracebackType
from typing import Any, Literal, Optional, TypeVar, Union

from rich import print as rprint
from rich.progress import (
    BarColumn,
    Console,
    Live,
    MofNCompleteColumn,
    Progress,
    ProgressColumn,
    SpinnerColumn,
    Task,
    Text,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)
from rq_geo_toolkit.rich_utils import FORCE_TERMINAL, VERBOSITY_MODE

__all__ = [
    "FORCE_TERMINAL",
    "VERBOSITY_MODE",
    "TaskProgressBar",
    "TaskProgressSpinner",
    "log_message",
    "show_total_elapsed_time",
]

TOTAL_STEPS = 33
ProgressType = TypeVar("ProgressType", bound=Progress)


def log_message(message: str) -> None:
    rprint(message)


def show_total_elapsed_time(elapsed_seconds: float) -> None:
    elapsed_time_formatted = str(timedelta(seconds=int(elapsed_seconds)))
    rprint(f"Finished operation in [progress.elapsed]{elapsed_time_formatted}")


class SpeedColumn(ProgressColumn):  # type: ignore[misc]
    def render(self, task: "Task") -> Text:
        if task.speed is None:
            return Text("")
        elif task.speed >= 1:
            return Text(f"{task.speed:.2f} it/s")
        else:
            return Text(f"{1/task.speed:.2f} s/it")  # noqa: FURB126


class TransientProgress(Progress):  # type: ignore[misc]
    def __init__(self, *columns: Union[str, ProgressColumn], live_obj: Live, **kwargs: Any) -> None:
        super().__init__(*columns, **kwargs)
        self.live = live_obj
        self.live._get_renderable = self.get_renderable

    def start(self) -> None:
        if not self.live.transient:
            super().start()

    def stop(self) -> None:
        if not self.live.transient:
            super().stop()

    def __enter__(self) -> Progress:
        self.start()
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self.stop()


class TaskProgressSpinner:
    def __init__(
        self,
        step_name: str,
        step_number: str,
        silent_mode: bool,
        transient_mode: bool,
        progress_cls: type[ProgressType],
        live_obj: Any,
        skip_step_number: bool = False,
    ):
        self.step_name = step_name
        self.step_number = step_number
        self.silent_mode = silent_mode
        self.transient_mode = transient_mode
        self.skip_step_number = skip_step_number
        self.progress = None
        self.progress_cls = progress_cls
        self.live_obj = live_obj

    def __enter__(self) -> None:
        if self.silent_mode:
            self.progress = None
        else:
            columns = [
                SpinnerColumn(),
                TextColumn(self.step_number),
                TextColumn("[progress.description]{task.description}"),
                TextColumn("•"),
                TimeElapsedColumn(),
            ]

            if self.skip_step_number:
                columns.pop(1)

            self.progress = self.progress_cls(
                *columns,
                live_obj=self.live_obj,
                transient=self.transient_mode,
            )

            self.progress.__enter__()  # type: ignore[attr-defined]
            self.progress.add_task(description=self.step_name, total=None)  # type: ignore[attr-defined]

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        if self.progress:
            self.progress.__exit__(exc_type, exc_val, exc_tb)

        self.progress = None


T = TypeVar("T")


class TaskProgressBar:
    def __init__(
        self,
        step_name: str,
        step_number: str,
        silent_mode: bool,
        transient_mode: bool,
        progress_cls: type[ProgressType],
        live_obj: Any,
        skip_step_number: bool = False,
    ):
        self.step_name = step_name
        self.step_number = step_number
        self.silent_mode = silent_mode
        self.transient_mode = transient_mode
        self.skip_step_number = skip_step_number
        self.progress = None
        self.progress_cls = progress_cls
        self.live_obj = live_obj

    def _create_progress(self) -> None:
        columns = [
            SpinnerColumn(),
            TextColumn(self.step_number),
            TextColumn(
                "[progress.description]{task.description}"
                " [progress.percentage]{task.percentage:>3.0f}%"
            ),
            BarColumn(),
            MofNCompleteColumn(),
            TextColumn("•"),
            TimeElapsedColumn(),
            TextColumn("<"),
            TimeRemainingColumn(),
            TextColumn("•"),
            SpeedColumn(),
        ]

        if self.skip_step_number:
            columns.pop(1)

        self.progress = self.progress_cls(
            *columns,
            live_obj=self.live_obj,
            transient=self.transient_mode,
            speed_estimate_period=1800,
        )

    def __enter__(self) -> "TaskProgressBar":
        if self.silent_mode:
            self.progress = None
        else:
            self._create_progress()
            self.progress.__enter__()  # type: ignore[attr-defined]

        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        if self.progress:
            self.progress.__exit__(exc_type, exc_val, exc_tb)

        self.progress = None

    def create_manual_bar(self, total: int) -> None:
        if self.progress:
            self.progress.add_task(description=self.step_name, total=total)

    def update_manual_bar(self, current_progress: int) -> None:
        if self.progress:
            self.progress.update(task_id=self.progress.task_ids[0], completed=current_progress)

    def track(self, iterable: Iterable[T]) -> Generator[T, None, None]:
        if self.progress is not None:
            for i in self.progress.track(list(iterable), description=self.step_name):
                yield i
        else:
            for i in iterable:
                yield i


class TaskProgressTracker:
    def __init__(
        self,
        verbosity_mode: VERBOSITY_MODE = "verbose",
        total_file_steps: int = 1,
        current_file_step: int = 1,
        debug: bool = False,
    ):
        self.verbosity_mode = verbosity_mode
        self.major_step_number: int = 0
        self.minor_step_number: Optional[int] = None
        self.total_files_steps = total_file_steps
        self.current_file_step = current_file_step
        self.live = None
        self.console = None
        self.debug = debug
        self.steps_times: dict[str, float] = {}
        self.start_time = time.time()

        if total_file_steps > 1:
            number_width = len(str(total_file_steps))
            self.major_steps_prefix = f"[{current_file_step: >{number_width}}/{total_file_steps}]"
        else:
            self.major_steps_prefix = ""

        if not self.verbosity_mode == "silent":
            self.console = Console(
                force_interactive=False if FORCE_TERMINAL else None,  # noqa: FURB110
                force_jupyter=False if FORCE_TERMINAL else None,  # noqa: FURB110
                force_terminal=True if FORCE_TERMINAL else None,  # noqa: FURB110
            )

    def reset_steps(self, current_file_step: int) -> None:
        self.major_step_number = 0
        self.minor_step_number = None

        self.current_file_step = current_file_step
        if self.total_files_steps > 1:
            number_width = len(str(self.total_files_steps))
            self.major_steps_prefix = (
                f"[{self.current_file_step: >{number_width}}/{self.total_files_steps}]"
            )
        else:
            self.major_steps_prefix = ""

    def is_new(self) -> bool:
        return self.major_step_number == 0 and self.minor_step_number is None

    def stop(self) -> None:
        if self.live:
            self.live.stop()
        if self.console and not self.console.is_interactive:
            self.console.print()

        if not self.verbosity_mode == "silent" and not self.is_new():
            end_time = time.time()
            elapsed_seconds = end_time - self.start_time
            show_total_elapsed_time(elapsed_seconds)

        if self.debug:
            self.steps_times["finish"] = round(time.time(), 2)
            rprint(f"Steps times: {json.dumps(self.steps_times)}")

    def _check_live_obj(self) -> None:
        if self.verbosity_mode == "silent":
            return
        if not self.live or not self.live._started:
            self.live = Live(
                console=self.console,
                auto_refresh=True,
                refresh_per_second=1,
                transient=self.verbosity_mode == "transient",
                redirect_stdout=True,
                redirect_stderr=True,
            )
            self.live.start(refresh=True)  # type: ignore[attr-defined]

    def get_spinner(
        self,
        step_name: str,
        next_step: Literal["major", "minor"] = "major",
        with_minor_step: bool = False,
    ) -> TaskProgressSpinner:
        self._parse_steps(next_step=next_step, with_minor_step=with_minor_step)
        self._check_live_obj()
        return TaskProgressSpinner(
            step_name=step_name,
            step_number=self.current_step_number,
            silent_mode=self.verbosity_mode == "silent",
            transient_mode=self.verbosity_mode == "transient",
            progress_cls=TransientProgress,
            live_obj=self.live,
        )

    def get_bar(
        self,
        step_name: str,
        next_step: Literal["major", "minor"] = "major",
        with_minor_step: bool = False,
    ) -> TaskProgressBar:
        self._parse_steps(next_step=next_step, with_minor_step=with_minor_step)
        self._check_live_obj()
        return TaskProgressBar(
            step_name=step_name,
            step_number=self.current_step_number,
            silent_mode=self.verbosity_mode == "silent",
            transient_mode=self.verbosity_mode == "transient",
            progress_cls=TransientProgress,
            live_obj=self.live,
        )

    def get_basic_bar(self, step_name: str) -> TaskProgressBar:
        self._check_live_obj()
        return TaskProgressBar(
            step_name=step_name,
            step_number="",
            silent_mode=self.verbosity_mode == "silent",
            transient_mode=self.verbosity_mode == "transient",
            skip_step_number=False,
            progress_cls=TransientProgress,
            live_obj=self.live,
        )

    def get_basic_spinner(self, step_name: str) -> TaskProgressSpinner:
        self._check_live_obj()
        return TaskProgressSpinner(
            step_name=step_name,
            step_number="",
            silent_mode=self.verbosity_mode == "silent",
            transient_mode=self.verbosity_mode == "transient",
            skip_step_number=False,
            progress_cls=TransientProgress,
            live_obj=self.live,
        )

    def _parse_steps(
        self, next_step: Literal["major", "minor"] = "major", with_minor_step: bool = False
    ) -> None:
        if next_step == "major":
            self.major_step_number += 1
            self.minor_step_number = 1 if with_minor_step else None
        elif next_step == "minor":
            if not self.minor_step_number:
                self.minor_step_number = 0
            self.minor_step_number += 1

        step_number = (
            str(self.major_step_number)
            if not self.minor_step_number
            else f"{self.major_step_number}.{self.minor_step_number}"
        )

        if self.debug:
            self.steps_times[step_number] = round(time.time(), 2)

        self.current_step_number = f"{self.major_steps_prefix}[{step_number: >4}/{TOTAL_STEPS}]"



"""Test big file pipeline."""

import shutil
from pathlib import Path

from parametrization import Parametrization as P
from pooch import get_logger as get_pooch_logger
from pooch import retrieve
from shapely import Polygon, box

from quackosm import PbfFileReader
from quackosm._osm_tags_filters import OsmTagsFilter


@P.parameters("extract_name", "geometry_filter", "tags_filter")  # type: ignore
@P.case(
    "Spain",
    "spain",
    box(-10.096436,35.777019,3.680420,44.040591),
    {"name:*": False},
)  # type: ignore
def test_big_file(extract_name: str, geometry_filter: Polygon, tags_filter: OsmTagsFilter) -> None:
    """Test if big file is working in a low memory environment."""
    files_dir = Path("files")
    shutil.rmtree(files_dir)
    file_name = files_dir / f"{extract_name}.osm.pbf"
    logger = get_pooch_logger()
    logger.setLevel("WARNING")
    retrieve(
        f"https://download.geofabrik.de/europe/{extract_name}-latest.osm.pbf",
        fname=f"{extract_name}.osm.pbf",
        path=files_dir,
        progressbar=False,
        known_hash=None,
    )

    reader = PbfFileReader(
        working_directory=files_dir,
        verbosity_mode="verbose",
        tags_filter=tags_filter,
        geometry_filter=geometry_filter,
    )
    # Reset rows_per_group value to test automatic downscaling
    reader.internal_rows_per_group = PbfFileReader.ROWS_PER_GROUP_MEMORY_CONFIG[24]
    reader.convert_pbf_to_parquet(pbf_path=file_name, ignore_cache=True)



"""
Geohash encoding/decoding and associated functions.

(c) Chris Veness 2014-2016 / MIT Licence https://www.movable-type.co.uk/scripts/geohash.html

(c) Anu Joy 2018 / MIT Licence
https://github.com/joyanujoy/geolib/blob/19c8e544e3ed6d39a13af6e96461346a778ae685/geolib/geohash.py#L50

http://en.wikipedia.org/wiki/Geohash
"""

from collections.abc import Generator

base32 = "0123456789bcdefghjkmnpqrstuvwxyz"


def _indexes(geohash: str) -> Generator[int, None, None]:
    if not geohash:
        raise ValueError("Invalid geohash")

    for char in geohash:
        try:
            yield base32.index(char)
        except ValueError:
            raise ValueError("Invalid geohash") from None


def geohash_bounds(geohash: str) -> tuple[float, float, float, float]:
    geohash = geohash.lower()

    even_bit = True
    lat_min = -90.0
    lat_max = 90.0
    lon_min = -180.0
    lon_max = 180.0

    # 5 bits for a char. So divide the decimal by power of 2, then AND 1
    # to get the binary bit - fast modulo operation.
    for index in _indexes(geohash):
        for n in range(4, -1, -1):
            bit = (index >> n) & 1
            if even_bit:
                # longitude
                lon_mid = (lon_min + lon_max) / 2
                if bit == 1:
                    lon_min = lon_mid
                else:
                    lon_max = lon_mid
            else:
                # latitude
                lat_mid = (lat_min + lat_max) / 2
                if bit == 1:
                    lat_min = lat_mid
                else:
                    lat_max = lat_mid
            even_bit = not even_bit

    return lon_min, lat_min, lon_max, lat_max



"""Tests related to OSM extracts."""

import datetime
from contextlib import nullcontext as does_not_raise
from typing import Any
from unittest import TestCase

import pandas as pd
import pytest
from dateutil.relativedelta import relativedelta
from parametrization import Parametrization as P
from pytest_mock import MockerFixture
from rich.console import Console
from shapely import box, from_wkt
from shapely.geometry.base import BaseGeometry

from quackosm._exceptions import (
    GeometryNotCoveredError,
    GeometryNotCoveredWarning,
    MissingOsmCacheWarning,
    OldOsmCacheWarning,
    OsmExtractIndexOutdatedWarning,
    OsmExtractMultipleMatchesError,
    OsmExtractZeroMatchesError,
)
from quackosm.geocode import geocode_to_geometry
from quackosm.osm_extracts import (
    OsmExtractSource,
    clear_osm_index_cache,
    display_available_extracts,
    find_smallest_containing_extracts,
    find_smallest_containing_extracts_total,
    get_extract_by_query,
)
from quackosm.osm_extracts.extract import (
    OpenStreetMapExtract,
    _get_full_file_name_function,
    _get_global_cache_file_path,
    _get_local_cache_file_path,
)
from quackosm.osm_extracts.geofabrik import _load_geofabrik_index

ut = TestCase()


@P.parameters("value")  # type: ignore
@P.case(
    "Base any",
    "any",
)  # type: ignore
@P.case(
    "Case insensitive any",
    "aNy",
)  # type: ignore
@P.case(
    "Base Geofabrik",
    "Geofabrik",
)  # type: ignore
@P.case(
    "Case insensitive Geofabrik",
    "GEOFABRIK",
)  # type: ignore
@P.case(
    "OSM fr without underscore",
    "osmfr",
)  # type: ignore
def test_proper_osm_extract_source(value: str) -> None:
    """Test if OsmExtractSource is parsed correctly."""
    OsmExtractSource(value)


def test_wrong_osm_extract_source():  # type: ignore
    """Test if cannot load incorrect OsmExtractSource."""
    with pytest.raises(ValueError):
        OsmExtractSource("test_source")


@P.parameters("source", "geometry", "expected_extract_id")  # type: ignore
@P.case(
    "Vatican - any",
    "any",
    from_wkt(
        "POLYGON ((12.450637854252449 41.904910802544634, 12.450637854252449 41.901790362263796,"
        " 12.455878610023916 41.901790362263796, 12.455878610023916 41.904910802544634,"
        " 12.450637854252449 41.904910802544634))"
    ),
    "osmfr_europe_vatican_city",
)  # type: ignore
@P.case(
    "Vatican - Geofabrik",
    "Geofabrik",
    from_wkt(
        "POLYGON ((12.450637854252449 41.904910802544634, 12.450637854252449 41.901790362263796,"
        " 12.455878610023916 41.901790362263796, 12.455878610023916 41.904910802544634,"
        " 12.450637854252449 41.904910802544634))"
    ),
    "Geofabrik_centro",
)  # type: ignore
@P.case(
    "London - any",
    "any",
    from_wkt(
        "POLYGON ((-0.1514787822171684 51.49843445562462, -0.1514787822171684 51.48926140694954,"
        " -0.1293785532031677 51.48926140694954, -0.1293785532031677 51.49843445562462,"
        " -0.1514787822171684 51.49843445562462))"
    ),
    "Geofabrik_greater-london",
)  # type: ignore
@P.case(
    "London - BBBike",
    "BBBike",
    from_wkt(
        "POLYGON ((-0.1514787822171684 51.49843445562462, -0.1514787822171684 51.48926140694954,"
        " -0.1293785532031677 51.48926140694954, -0.1293785532031677 51.49843445562462,"
        " -0.1514787822171684 51.49843445562462))"
    ),
    "BBBike_London",
)  # type: ignore
@P.case(
    "Vancouver - any",
    "any",
    from_wkt(
        "POLYGON ((-123.15817514738828 49.29493379142323, -123.15817514738828 49.23700029433431,"
        " -123.07449492760279 49.23700029433431, -123.07449492760279 49.29493379142323,"
        " -123.15817514738828 49.29493379142323))"
    ),
    "BBBike_Vancouver",
)  # type: ignore
@P.case(
    "Vancouver - OSM.fr",
    "osmfr",
    from_wkt(
        "POLYGON ((-123.15817514738828 49.29493379142323, -123.15817514738828 49.23700029433431,"
        " -123.07449492760279 49.23700029433431, -123.07449492760279 49.29493379142323,"
        " -123.15817514738828 49.29493379142323))"
    ),
    "osmfr_north-america_canada_british_columbia",
)  # type: ignore
def test_single_smallest_extract(
    source: str, geometry: BaseGeometry, expected_extract_id: str
) -> None:
    """Test if extracts matching works correctly for geometries within borders."""
    extracts = find_smallest_containing_extracts(geometry, source)
    assert len(extracts) == 1
    assert extracts[0].id == expected_extract_id, f"{extracts[0].id} vs {expected_extract_id}"


@P.parameters(
    "source", "geometry", "geometry_coverage_iou_threshold", "expected_extract_file_names"
)  # type: ignore
@P.case(
    "Andorra bbox, any, iou default",
    "any",
    from_wkt(
        "POLYGON ((1.382599544073372 42.67676873293743, 1.382599544073372 42.40065303248514,"
        " 1.8092269635579328 42.40065303248514, 1.8092269635579328 42.67676873293743,"
        " 1.382599544073372 42.67676873293743))"
    ),
    0.01,
    [
        "osmfr_europe_spain_catalunya_lleida",
        "osmfr_europe_france_midi_pyrenees_ariege",
        "osmfr_europe_france_languedoc_roussillon_pyrenees_orientales",
        "geofabrik_europe_andorra",
    ],
)  # type: ignore
@P.case(
    "Andorra bbox, any, iou 0",
    "any",
    from_wkt(
        "POLYGON ((1.382599544073372 42.67676873293743, 1.382599544073372 42.40065303248514,"
        " 1.8092269635579328 42.40065303248514, 1.8092269635579328 42.67676873293743,"
        " 1.382599544073372 42.67676873293743))"
    ),
    0,
    [
        "osmfr_europe_spain_catalunya_lleida",
        "osmfr_europe_spain_catalunya_girona",
        "osmfr_europe_france_midi_pyrenees_ariege",
        "osmfr_europe_france_languedoc_roussillon_pyrenees_orientales",
        "geofabrik_europe_andorra",
    ],
)  # type: ignore
@P.case(
    "Andorra geocode, geofabrik, iou default",
    "geofabrik",
    geocode_to_geometry("Andorra"),
    0.01,
    ["geofabrik_europe_andorra"],
)  # type: ignore
@P.case(
    "Andorra geocode, osmfr, iou 0",
    "osmfr",
    geocode_to_geometry("Andorra"),
    0,
    ["osmfr_europe"],
)  # type: ignore
def test_multiple_smallest_extracts(
    source: str,
    geometry: BaseGeometry,
    geometry_coverage_iou_threshold: float,
    expected_extract_file_names: list[str],
) -> None:
    """Test if extracts matching works correctly for geometries between borders."""
    extracts = find_smallest_containing_extracts(
        geometry, source, geometry_coverage_iou_threshold=geometry_coverage_iou_threshold
    )
    ut.assertListEqual([extract.file_name for extract in extracts], expected_extract_file_names)


@pytest.mark.parametrize(
    "expectation,allow_uncovered_geometry,geometry_coverage_iou_threshold",
    [
        (pytest.raises(GeometryNotCoveredError), False, 0.01),
        (pytest.raises(ValueError), False, -0.1),
        (pytest.raises(ValueError), False, 1.2),
        (pytest.raises(ValueError), True, 1.2),
        (pytest.warns(GeometryNotCoveredWarning), True, 0.01),
    ],
)  # type: ignore
def test_uncovered_geometry_extract(
    expectation: Any, allow_uncovered_geometry: bool, geometry_coverage_iou_threshold: float
) -> None:
    """Test if raises errors as expected when geometry can't be covered."""
    with expectation:
        geometry = from_wkt(
            "POLYGON ((-43.064 29.673, -43.064 29.644, -43.017 29.644,"
            " -43.017 29.673, -43.064 29.673))"
        )
        find_smallest_containing_extracts_total(
            geometry=geometry,
            allow_uncovered_geometry=allow_uncovered_geometry,
            geometry_coverage_iou_threshold=geometry_coverage_iou_threshold,
        )


def test_proper_cache_saving() -> None:
    """Test if file is saved in cache properly."""
    save_path = _get_global_cache_file_path(OsmExtractSource.geofabrik)
    loaded_index = _load_geofabrik_index()
    assert save_path.exists()
    assert len(loaded_index.columns) == 7


def test_wrong_cached_index() -> None:
    """Test if cached file with missing columns is redownloaded again."""
    save_path = _get_global_cache_file_path(OsmExtractSource.geofabrik)
    column_to_remove = "id"

    # load index first time
    first_index = _load_geofabrik_index()

    # remove the column and replace the file
    first_index.drop(columns=[column_to_remove]).to_file(save_path, driver="GeoJSON")

    with pytest.warns(OsmExtractIndexOutdatedWarning):
        # load index again
        second_index = _load_geofabrik_index()

    assert column_to_remove in second_index.columns


def test_proper_full_name() -> None:
    """Test if full names for extracts are properly generated."""
    test_index = pd.DataFrame({"id": ["1", "2"], "name": ["one", "two"], "parent": ["x", "1"]})
    prepared_function = _get_full_file_name_function(test_index)
    assert prepared_function("2") == "x_one_two"


@P.parameters("query", "source", "expectation", "matched_id", "exception_values")  # type: ignore
@P.case(
    "Proper full file name Geobfabrik",
    "geofabrik_north-america_us",
    "geofabrik",
    does_not_raise(),
    "Geofabrik_us",
    [],
)  # type: ignore
@P.case(
    "Proper full file name any",
    "geofabrik_north-america_us",
    "any",
    does_not_raise(),
    "Geofabrik_us",
    [],
)  # type: ignore
@P.case(
    "Proper name bbbike",
    "London",
    "BBBike",
    does_not_raise(),
    "BBBike_London",
    [],
)  # type: ignore
@P.case(
    "Proper name bbbike - upper case",
    "LONDON",
    "BBBike",
    does_not_raise(),
    "BBBike_London",
    [],
)  # type: ignore
@P.case(
    "Proper name any - with whitespaces",
    "   tete  ",
    "any",
    does_not_raise(),
    "osmfr_africa_mozambique_tete",
    [],
)  # type: ignore
@P.case(
    "Wrong query multiple matches - single source",
    "northeast",
    "any",
    pytest.raises(OsmExtractMultipleMatchesError),
    "",
    [
        "osmfr_north-america_us-midwest_illinois_northeast",
        "osmfr_north-america_us-south_florida_northeast",
        "osmfr_north-america_us-south_georgia_northeast",
        "osmfr_north-america_us-south_north-carolina_northeast",
        "osmfr_north-america_us-west_colorado_northeast",
        "osmfr_south-america_brazil_northeast",
    ],
)  # type: ignore
@P.case(
    "Wrong query multiple matches - multiple sources",
    "asia",
    "any",
    pytest.raises(OsmExtractMultipleMatchesError),
    "",
    ["geofabrik_asia", "osmfr_asia"],
)  # type: ignore
@P.case(
    "Wrong query zero matches with suggestions - north",
    "nrth",
    "any",
    pytest.raises(OsmExtractZeroMatchesError),
    "",
    [
        "osmfr_north-america_us-midwest_illinois_north",
        "osmfr_north-america_us-south_texas_north",
        "osmfr_south-america_brazil_north",
    ],
)  # type: ignore
@P.case(
    "Wrong query zero matches with suggestions - prlnd",
    "prlnd",
    "any",
    pytest.raises(OsmExtractZeroMatchesError),
    "",
    [
        "bbbike_portland",
        "osmfr_europe_poland",
        "geofabrik_europe_poland",
    ],
)  # type: ignore
@P.case(
    "Wrong query zero matches without suggestions",
    "empty_extract",
    "any",
    pytest.raises(OsmExtractZeroMatchesError),
    "",
    [],
)  # type: ignore
def test_extracts_finding(
    query: str, source: str, expectation: Any, matched_id: str, exception_values: list[str]
) -> None:
    """Test if extracts finding by name works."""
    with expectation as exception_info:
        extract = get_extract_by_query(query, source)
        # if properly found - check id
        assert extract.id == matched_id

    # if threw exception - check resulting arrays
    if exception_info is not None:
        ut.assertListEqual(exception_info.value.matching_full_names, exception_values)


@pytest.mark.parametrize(
    "use_full_names",
    [False, True],
)  # type: ignore
@pytest.mark.parametrize(
    "osm_source",
    list(OsmExtractSource),
)  # type: ignore
def test_extracts_tree_printing(
    capfd: Any, mocker: MockerFixture, osm_source: OsmExtractSource, use_full_names: bool
) -> None:
    """Test if displaying available extracts works."""
    mocker.patch("rich.get_console", return_value=Console(width=999))
    display_available_extracts(osm_source, use_full_names)
    output, error_output = capfd.readouterr()

    assert len(output) > 0

    osm_sources_without_any = [src for src in OsmExtractSource if src != OsmExtractSource.any]

    if osm_source == OsmExtractSource.any:
        assert output.startswith("All extracts")
        assert all(src.value in output for src in osm_sources_without_any)
    else:
        assert output.startswith(osm_source.value)

    if use_full_names:
        lines = output.lower().split("\n")

        assert all(
            any(src.value.lower() in line for src in osm_sources_without_any)
            for line in lines
            if len(line.strip()) > 0 and line != "all extracts"
        )

    assert error_output == ""


def test_generate_index_warning(mocker: MockerFixture) -> None:
    """Test if index generation results in warning."""
    extract_source = OsmExtractSource.bbbike
    global_path = _get_global_cache_file_path(extract_source)
    local_path = _get_local_cache_file_path(extract_source)

    move_global_path = global_path.exists()
    move_local_path = local_path.exists()

    if move_global_path:
        global_moved_path = global_path.with_name("bbbike_index_moved.geojson")
        global_path.rename(global_moved_path)

    if move_local_path:
        local_moved_path = local_path.with_name("bbbike_index_moved.geojson")
        local_path.rename(local_moved_path)

    try:
        mocker.patch(
            "quackosm.osm_extracts.bbbike._iterate_bbbike_index",
            return_value=[
                OpenStreetMapExtract(
                    id="bbbike_test",
                    name="test",
                    parent="bbbike",
                    url="test_url",
                    geometry=box(0, 0, 1, 1),
                )
            ],
        )
        mocker.patch(
            "quackosm.osm_extracts.bbbike.BBBIKE_INDEX_GDF", new=None
        )
        with pytest.warns(MissingOsmCacheWarning):
            display_available_extracts(source=extract_source)

    finally:
        if move_global_path:
            global_path.unlink(missing_ok=True)
            global_moved_path.rename(global_path)
            global_moved_path.unlink(missing_ok=True)

        if move_local_path:
            local_path.unlink(missing_ok=True)
            local_moved_path.rename(local_path)
            local_moved_path.unlink(missing_ok=True)


def test_old_index_warning(mocker: MockerFixture) -> None:
    """Test if old index results in warning."""
    extract_source = OsmExtractSource.bbbike

    mocker.patch(
        "quackosm.osm_extracts.bbbike._iterate_bbbike_index",
        return_value=[
            OpenStreetMapExtract(
                id="bbbike_test",
                name="test",
                parent="bbbike",
                url="test_url",
                geometry=box(0, 0, 1, 1),
            )
        ],
    )
    mocker.patch(
        "quackosm.osm_extracts.extract._get_file_creation_date",
        return_value=datetime.datetime.now() - relativedelta(years=1, days=1),
    )
    mocker.patch(
        "quackosm.osm_extracts.bbbike.BBBIKE_INDEX_GDF", new=None
    )

    with pytest.warns(OldOsmCacheWarning):
        display_available_extracts(source=extract_source)

def test_cache_clearing() -> None:
    """Test if cache clearing works."""
    extract_source = OsmExtractSource.bbbike
    global_path = _get_global_cache_file_path(extract_source)
    local_path = _get_local_cache_file_path(extract_source)

    move_global_path = global_path.exists()
    move_local_path = local_path.exists()

    if move_global_path:
        global_moved_path = global_path.with_name("bbbike_index_moved.geojson")
        global_path.rename(global_moved_path)

    if move_local_path:
        local_moved_path = local_path.with_name("bbbike_index_moved.geojson")
        local_path.rename(local_moved_path)

    clear_osm_index_cache(extract_source)

    assert not global_path.exists()
    assert not local_path.exists()

    if move_global_path:
        global_moved_path.rename(global_path)
        global_moved_path.unlink(missing_ok=True)

    if move_local_path:
        local_moved_path.rename(local_path)
        local_moved_path.unlink(missing_ok=True)



"""
OpenStreetMap extracts.

This module contains iterators for publically available OpenStreetMap `*.osm.pbf` files
repositories.
"""

import difflib
import warnings
from collections.abc import Iterable
from functools import partial
from math import ceil
from multiprocessing import cpu_count
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union, overload

import geopandas as gpd
from pandas.util._decorators import deprecate
from pooch import get_logger as get_pooch_logger
from pooch import retrieve
from rich import get_console
from rich import print as rprint
from rq_geo_toolkit._geopandas_api_version import GEOPANDAS_NEW_API
from shapely.geometry.base import BaseGeometry, BaseMultipartGeometry
from tqdm.contrib.concurrent import process_map

from quackosm._exceptions import (
    GeometryNotCoveredError,
    GeometryNotCoveredWarning,
    OsmExtractMultipleMatchesError,
    OsmExtractZeroMatchesError,
)
from quackosm._rich_progress import FORCE_TERMINAL
from quackosm.osm_extracts.bbbike import _get_bbbike_index
from quackosm.osm_extracts.extract import (
    OpenStreetMapExtract,
    OsmExtractSource,
    clear_osm_index_cache,
)
from quackosm.osm_extracts.extracts_tree import get_available_extracts_as_rich_tree
from quackosm.osm_extracts.geofabrik import _get_geofabrik_index
from quackosm.osm_extracts.osm_fr import _get_openstreetmap_fr_index

if TYPE_CHECKING:  # pragma: no cover
    import pandas as pd

__all__ = [
    "download_extracts_pbf_files",
    "find_smallest_containing_extracts_total",
    "find_smallest_containing_geofabrik_extracts",
    "find_smallest_containing_openstreetmap_fr_extracts",
    "find_smallest_containing_bbbike_extracts",
    "clear_osm_index_cache",
    "get_extract_by_query",
    "download_extract_by_query",
    "display_available_extracts",
    "OsmExtractSource",
]


def download_extracts_pbf_files(
    extracts: list[OpenStreetMapExtract], download_directory: Path, progressbar: bool = True
) -> list[Path]:
    """
    Download OSM extracts as PBF files.

    Args:
        extracts (list[OpenStreetMapExtract]): List of extracts to download.
        download_directory (Path): Directory where PBF files should be saved.
        progressbar (bool, optional): Show progress bar. Defaults to True.

    Returns:
        list[Path]: List of downloaded file paths.
    """
    downloaded_extracts_paths = []
    logger = get_pooch_logger()
    logger.setLevel("WARNING")
    for extract in extracts:
        file_path = retrieve(
            extract.url,
            fname=f"{extract.file_name}.osm.pbf",
            path=download_directory,
            progressbar=progressbar and not FORCE_TERMINAL,
            known_hash=None,
        )
        downloaded_extracts_paths.append(Path(file_path))
    return downloaded_extracts_paths


OSM_EXTRACT_SOURCE_INDEX_FUNCTION = {
    OsmExtractSource.bbbike: _get_bbbike_index,
    OsmExtractSource.geofabrik: _get_geofabrik_index,
    OsmExtractSource.osm_fr: _get_openstreetmap_fr_index,
}


def _get_combined_index() -> gpd.GeoDataFrame:
    combined_index = gpd.pd.concat(
        [get_index_function() for get_index_function in OSM_EXTRACT_SOURCE_INDEX_FUNCTION.values()]
    )
    combined_index.sort_values(by="area", ignore_index=True, inplace=True)
    return combined_index


@overload
def get_extract_by_query(query: str) -> OpenStreetMapExtract: ...


@overload
def get_extract_by_query(
    query: str, source: Union[OsmExtractSource, str]
) -> OpenStreetMapExtract: ...


def get_extract_by_query(
    query: str, source: Union[OsmExtractSource, str] = "any"
) -> OpenStreetMapExtract:
    """
    Find an OSM extract by name.

    Args:
        query (str): Query to search for a particular extract.
        source (Union[OsmExtractSource, str]): OSM source name. Can be one of: 'any', 'Geofabrik',
            'BBBike', 'OSM_fr'. Defaults to 'any'.

    Returns:
        OpenStreetMapExtract: Found extract.
    """
    try:
        source_enum = OsmExtractSource(source)
        index = OSM_EXTRACT_SOURCE_INDEX_FUNCTION.get(source_enum, _get_combined_index)()

        matching_index_row: pd.Series = None

        file_name_matched_rows = (index["file_name"].str.lower() == query.lower().strip()) | (
            index["file_name"].str.replace("_", " ").str.lower()
            == query.lower().replace("_", " ").strip()
        )
        extract_name_matched_rows = (index["name"].str.lower() == query.lower().strip()) | (
            index["name"].str.replace("_", " ").str.lower()
            == query.lower().replace("_", " ").strip()
        )

        # full file name matched
        if sum(file_name_matched_rows) == 1:
            matching_index_row = index[file_name_matched_rows].iloc[0]
        # single name matched
        elif sum(extract_name_matched_rows) == 1:
            matching_index_row = index[extract_name_matched_rows].iloc[0]
        # multiple names matched
        elif extract_name_matched_rows.any():
            matching_rows = index[extract_name_matched_rows]
            matching_full_names = sorted(matching_rows["file_name"])
            full_names = ", ".join(f'"{full_name}"' for full_name in matching_full_names)

            raise OsmExtractMultipleMatchesError(
                f'Multiple extracts matched by query "{query.strip()}".\n'
                f"Matching extracts full names: {full_names}.",
                matching_full_names=matching_full_names,
            )
        # zero names matched
        elif not extract_name_matched_rows.any():
            matching_full_names = []
            suggested_query_names = difflib.get_close_matches(
                query.lower(), index["name"].str.lower().unique(), n=5, cutoff=0.7
            )

            if suggested_query_names:
                for suggested_query_name in suggested_query_names:
                    found_extracts = index[index["name"].str.lower() == suggested_query_name]
                    matching_full_names.extend(found_extracts["file_name"])
                full_names = ", ".join(matching_full_names)
                full_names = ", ".join(f'"{full_name}"' for full_name in matching_full_names)
                exception_message = (
                    f'Zero extracts matched by query "{query}".\n'
                    f"Found full names close to query: {full_names}."
                )
            else:
                exception_message = (
                    f'Zero extracts matched by query "{query}".\n'
                    "Zero close matches have been found."
                )

            raise OsmExtractZeroMatchesError(
                exception_message,
                matching_full_names=matching_full_names,
            )

        return OpenStreetMapExtract(
            id=matching_index_row["id"],
            name=matching_index_row["name"],
            parent=matching_index_row["parent"],
            url=matching_index_row["url"],
            geometry=matching_index_row["geometry"],
            file_name=matching_index_row["file_name"],
        )

    except ValueError as ex:
        raise ValueError(f"Unknown OSM extracts source: {source}.") from ex


@overload
def download_extract_by_query(
    query: str, *, download_directory: Union[str, Path] = "files", progressbar: bool = True
) -> Path: ...


@overload
def download_extract_by_query(
    query: str,
    source: Union[OsmExtractSource, str],
    *,
    download_directory: Union[str, Path] = "files",
    progressbar: bool = True,
) -> Path: ...


def download_extract_by_query(
    query: str,
    source: Union[OsmExtractSource, str] = "any",
    download_directory: Union[str, Path] = "files",
    progressbar: bool = True,
) -> Path:
    """
    Download an OSM extract by name.

    Args:
        query (str): Query to search for a particular extract.
        source (Union[OsmExtractSource, str]): OSM source name. Can be one of: 'any', 'Geofabrik',
            'BBBike', 'OSM_fr'. Defaults to 'any'.
        download_directory (Union[str, Path], optional): Directory where the file should be
            downloaded. Defaults to "files".
        progressbar (bool, optional): Show progress bar. Defaults to True.

    Returns:
        Path: Path to the downloaded OSM extract.
    """
    matching_extract = get_extract_by_query(query, source)
    return download_extracts_pbf_files([matching_extract], Path(download_directory), progressbar)[0]


def display_available_extracts(
    source: Union[OsmExtractSource, str],
    use_full_names: bool = True,
    use_pager: bool = False,
) -> None:
    """
    Display all available OSM extracts in the form of a tree.

    Output will be printed to the console.

    Args:
        source (Union[OsmExtractSource, str]): Source for which extracts should be displayed.
        use_full_names (bool, optional): Whether to display full name, or short name of the extract.
            Full name contains all parents of the extract. Defaults to `True`.
        use_pager (bool, optional): Whether to display long output using Rich pager
            or just print to output. Defaults to `False`.

    Raises:
        ValueError: If provided source value cannot be parsed to OsmExtractSource.
    """
    try:
        source_enum = OsmExtractSource(source)
        tree = get_available_extracts_as_rich_tree(
            source_enum, OSM_EXTRACT_SOURCE_INDEX_FUNCTION, use_full_names
        )
        if not use_pager:
            rprint(tree)
        else:
            console = get_console()
            with console.pager():
                console.print(tree)
    except ValueError as ex:
        raise ValueError(f"Unknown OSM extracts source: {source}.") from ex


def find_smallest_containing_extracts_total(
    geometry: Union[BaseGeometry, BaseMultipartGeometry],
    geometry_coverage_iou_threshold: float = 0.01,
    allow_uncovered_geometry: bool = False,
) -> list[OpenStreetMapExtract]:
    """
    Find smallest extracts from all OSM extract indexes that contains given polygon.

    Iterates all indexes and finds smallest extracts that covers a given geometry.

    Extracts are selected based on the highest value of the Intersection over Union metric with
    geometry. Some extracts might be discarded because of low IoU metric value leaving some parts
    of the geometry uncovered.

    Args:
        geometry (Union[BaseGeometry, BaseMultipartGeometry]): Geometry to be covered.
        geometry_coverage_iou_threshold (float): Minimal value of the Intersection over Union metric
            for selecting the matching OSM extracts. Is best matching extract has value lower than
            the threshold, it is discarded (except the first one). Has to be in range between
            0 and 1. Value of 0 will allow every intersected extract, value of 1 will only allow
            extracts that match the geometry exactly. Defaults to 0.01.
        allow_uncovered_geometry (bool): Suppress an error if some geometry parts aren't covered
            by any OSM extract. Defaults to `False`.

    Returns:
        List[OpenStreetMapExtract]: List of extracts name, URL to download it and boundary polygon.
    """
    return _find_smallest_containing_extracts(
        geometry=geometry,
        polygons_index_gdf=_get_combined_index(),
        geometry_coverage_iou_threshold=geometry_coverage_iou_threshold,
        allow_uncovered_geometry=allow_uncovered_geometry,
    )


def find_smallest_containing_geofabrik_extracts(
    geometry: Union[BaseGeometry, BaseMultipartGeometry],
    geometry_coverage_iou_threshold: float = 0.01,
    allow_uncovered_geometry: bool = False,
) -> list[OpenStreetMapExtract]:
    """
    Find smallest extracts from Geofabrik that contains given geometry.

    Iterates a geofabrik index and finds smallest extracts that covers a given geometry.

    Extracts are selected based on the highest value of the Intersection over Union metric with
    geometry. Some extracts might be discarded because of low IoU metric value leaving some parts
    of the geometry uncovered.

    Args:
        geometry (Union[BaseGeometry, BaseMultipartGeometry]): Geometry to be covered.
        geometry_coverage_iou_threshold (float): Minimal value of the Intersection over Union metric
            for selecting the matching OSM extracts. Is best matching extract has value lower than
            the threshold, it is discarded (except the first one). Has to be in range between
            0 and 1. Value of 0 will allow every intersected extract, value of 1 will only allow
            extracts that match the geometry exactly. Defaults to 0.01.
        allow_uncovered_geometry (bool): Suppress an error if some geometry parts aren't covered
            by any OSM extract. Defaults to `False`.

    Returns:
        List[OpenStreetMapExtract]: List of extracts name, URL to download it and boundary polygon.
    """
    return _find_smallest_containing_extracts(
        geometry=geometry,
        polygons_index_gdf=OSM_EXTRACT_SOURCE_INDEX_FUNCTION[OsmExtractSource.geofabrik](),
        geometry_coverage_iou_threshold=geometry_coverage_iou_threshold,
        allow_uncovered_geometry=allow_uncovered_geometry,
    )


def find_smallest_containing_openstreetmap_fr_extracts(
    geometry: Union[BaseGeometry, BaseMultipartGeometry],
    geometry_coverage_iou_threshold: float = 0.01,
    allow_uncovered_geometry: bool = False,
) -> list[OpenStreetMapExtract]:
    """
    Find smallest extracts from OpenStreetMap.fr that contains given polygon.

    Iterates an osm.fr index and finds smallest extracts that covers a given geometry.

    Extracts are selected based on the highest value of the Intersection over Union metric with
    geometry. Some extracts might be discarded because of low IoU metric value leaving some parts
    of the geometry uncovered.

    Args:
        geometry (Union[BaseGeometry, BaseMultipartGeometry]): Geometry to be covered.
        geometry_coverage_iou_threshold (float): Minimal value of the Intersection over Union metric
            for selecting the matching OSM extracts. Is best matching extract has value lower than
            the threshold, it is discarded (except the first one). Has to be in range between
            0 and 1. Value of 0 will allow every intersected extract, value of 1 will only allow
            extracts that match the geometry exactly. Defaults to 0.01.
        allow_uncovered_geometry (bool): Suppress an error if some geometry parts aren't covered
            by any OSM extract. Defaults to `False`.

    Returns:
        List[OpenStreetMapExtract]: List of extracts name, URL to download it and boundary polygon.
    """
    return _find_smallest_containing_extracts(
        geometry=geometry,
        polygons_index_gdf=OSM_EXTRACT_SOURCE_INDEX_FUNCTION[OsmExtractSource.osm_fr](),
        geometry_coverage_iou_threshold=geometry_coverage_iou_threshold,
        allow_uncovered_geometry=allow_uncovered_geometry,
    )


def find_smallest_containing_bbbike_extracts(
    geometry: Union[BaseGeometry, BaseMultipartGeometry],
    geometry_coverage_iou_threshold: float = 0.01,
    allow_uncovered_geometry: bool = False,
) -> list[OpenStreetMapExtract]:
    """
    Find smallest extracts from BBBike that contains given polygon.

    Iterates an BBBike index and finds smallest extracts that covers a given geometry.

    Extracts are selected based on the highest value of the Intersection over Union metric with
    geometry. Some extracts might be discarded because of low IoU metric value leaving some parts
    of the geometry uncovered.

    Args:
        geometry (Union[BaseGeometry, BaseMultipartGeometry]): Geometry to be covered.
        geometry_coverage_iou_threshold (float): Minimal value of the Intersection over Union metric
            for selecting the matching OSM extracts. Is best matching extract has value lower than
            the threshold, it is discarded (except the first one). Has to be in range between
            0 and 1. Value of 0 will allow every intersected extract, value of 1 will only allow
            extracts that match the geometry exactly. Defaults to 0.01.
        allow_uncovered_geometry (bool): Suppress an error if some geometry parts aren't covered
            by any OSM extract. Defaults to `False`.

    Returns:
        List[OpenStreetMapExtract]: List of extracts name, URL to download it and boundary polygon.
    """
    return _find_smallest_containing_extracts(
        geometry=geometry,
        polygons_index_gdf=OSM_EXTRACT_SOURCE_INDEX_FUNCTION[OsmExtractSource.bbbike](),
        geometry_coverage_iou_threshold=geometry_coverage_iou_threshold,
        allow_uncovered_geometry=allow_uncovered_geometry,
    )


def find_smallest_containing_extracts(
    geometry: Union[BaseGeometry, BaseMultipartGeometry],
    source: Union[OsmExtractSource, str],
    geometry_coverage_iou_threshold: float = 0.01,
    allow_uncovered_geometry: bool = False,
) -> list[OpenStreetMapExtract]:
    """
    Find smallest extracts from a given OSM source that contains given polygon.

    Iterates an OSM source index and finds smallest extracts that covers a given geometry.

    Extracts are selected based on the highest value of the Intersection over Union metric with
    geometry. Some extracts might be discarded because of low IoU metric value leaving some parts
    of the geometry uncovered.

    Args:
        geometry (Union[BaseGeometry, BaseMultipartGeometry]): Geometry to be covered.
        source (Union[OsmExtractSource, str]): OSM source name. Can be one of: 'any', 'Geofabrik',
            'BBBike', 'OSMfr'.
        geometry_coverage_iou_threshold (float): Minimal value of the Intersection over Union metric
            for selecting the matching OSM extracts. Is best matching extract has value lower than
            the threshold, it is discarded (except the first one). Has to be in range between
            0 and 1. Value of 0 will allow every intersected extract, value of 1 will only allow
            extracts that match the geometry exactly. Defaults to 0.01.
        allow_uncovered_geometry (bool): Suppress an error if some geometry parts aren't covered
            by any OSM extract. Defaults to `False`.

    Returns:
        List[OpenStreetMapExtract]: List of extracts name, URL to download it and boundary polygon.
    """
    try:
        source_enum = OsmExtractSource(source)
        index = OSM_EXTRACT_SOURCE_INDEX_FUNCTION.get(source_enum, _get_combined_index)()
        return _find_smallest_containing_extracts(
            geometry=geometry,
            polygons_index_gdf=index,
            geometry_coverage_iou_threshold=geometry_coverage_iou_threshold,
            allow_uncovered_geometry=allow_uncovered_geometry,
        )
    except ValueError as ex:
        raise ValueError(f"Unknown OSM extracts source: {source}.") from ex


def _find_smallest_containing_extracts(
    geometry: Union[BaseGeometry, BaseMultipartGeometry],
    polygons_index_gdf: gpd.GeoDataFrame,
    num_of_multiprocessing_workers: int = -1,
    multiprocessing_activation_threshold: Optional[int] = None,
    geometry_coverage_iou_threshold: float = 0.01,
    allow_uncovered_geometry: bool = False,
) -> list[OpenStreetMapExtract]:
    """
    Find smallest set of extracts covering a given geometry.

    Iterates a provided extracts index and searches for a smallest set that cover a given geometry.
    It's not guaranteed that this set will be the smallest and there will be no overlaps.

    Extracts are selected based on the highest value of the Intersection over Union metric with
    geometry. Some extracts might be discarded because of low IoU metric value leaving some parts
    of the geometry uncovered.

    Geometry will be flattened into singluar geometries if it's `BaseMultipartGeometry`.

    Args:
        geometry (Union[BaseGeometry, BaseMultipartGeometry]): Geometry to be covered.
        polygons_index_gdf (gpd.GeoDataFrame): Index of available extracts.
        num_of_multiprocessing_workers (int, optional): Number of workers used for multiprocessing.
            Defaults to -1 which results in a total number of available cpu threads.
            `0` and `1` values disable multiprocessing.
            Similar to `n_jobs` parameter from `scikit-learn` library.
        multiprocessing_activation_threshold (int, optional): Number of gometries required to start
            processing on multiple processes. Activating multiprocessing for a small
            amount of points might not be feasible. Defaults to 100.
        geometry_coverage_iou_threshold (float): Minimal value of the Intersection over Union metric
            for selecting the matching OSM extracts. Is best matching extract has value lower than
            the threshold, it is discarded (except the first one). Has to be in range between
            0 and 1. Value of 0 will allow every intersected extract, value of 1 will only allow
            extracts that match the geometry exactly. Defaults to 0.01.
        allow_uncovered_geometry (bool): Suppress an error if some geometry parts aren't covered
            by any OSM extract. Defaults to `False`.

    Returns:
        List[OpenStreetMapExtract]: List of extracts covering a given geometry.
    """
    if num_of_multiprocessing_workers == 0:
        num_of_multiprocessing_workers = 1
    elif num_of_multiprocessing_workers < 0:
        num_of_multiprocessing_workers = cpu_count()

    if not multiprocessing_activation_threshold:
        multiprocessing_activation_threshold = 100

    unique_extracts_ids: set[str] = set()

    geometries = _flatten_geometry(geometry)

    total_polygons = len(geometries)

    if (
        num_of_multiprocessing_workers > 1
        and total_polygons >= multiprocessing_activation_threshold
    ):
        find_extracts_func = partial(
            _find_smallest_containing_extracts_for_single_geometry,
            polygons_index_gdf=polygons_index_gdf,
            geometry_coverage_iou_threshold=geometry_coverage_iou_threshold,
            allow_uncovered_geometry=allow_uncovered_geometry,
        )

        for extract_ids_list in process_map(
            find_extracts_func,
            geometries,
            desc="Finding matching extracts",
            max_workers=num_of_multiprocessing_workers,
            chunksize=ceil(total_polygons / (4 * num_of_multiprocessing_workers)),
            disable=FORCE_TERMINAL,
        ):
            unique_extracts_ids.update(extract_ids_list)
    else:
        for sub_geometry in geometries:
            unique_extracts_ids.update(
                _find_smallest_containing_extracts_for_single_geometry(
                    geometry=sub_geometry,
                    polygons_index_gdf=polygons_index_gdf,
                    geometry_coverage_iou_threshold=geometry_coverage_iou_threshold,
                    allow_uncovered_geometry=allow_uncovered_geometry,
                )
            )

    extracts_filtered = _filter_extracts(
        geometry,
        unique_extracts_ids,
        polygons_index_gdf,
        num_of_multiprocessing_workers,
        multiprocessing_activation_threshold,
    )

    return extracts_filtered


def _find_smallest_containing_extracts_for_single_geometry(
    geometry: BaseGeometry,
    polygons_index_gdf: gpd.GeoDataFrame,
    geometry_coverage_iou_threshold: float = 0.01,
    allow_uncovered_geometry: bool = False,
) -> set[str]:
    """
    Find smallest set of extracts covering a given singular geometry.

    Extracts are selected based on the highest value of the Intersection over Union metric with
    geometry. Some extracts might be discarded because of low IoU metric value leaving some parts
    of the geometry uncovered.

    Args:
        geometry (BaseGeometry): Geometry to be covered.
        polygons_index_gdf (gpd.GeoDataFrame): Index of available extracts.
        geometry_coverage_iou_threshold (float): Minimal value of the Intersection over Union metric
            for selecting the matching OSM extracts. Is best matching extract has value lower than
            the threshold, it is discarded (except the first one). Has to be in range between
            0 and 1. Value of 0 will allow every intersected extract, value of 1 will only allow
            extracts that match the geometry exactly. Defaults to 0.01.
        allow_uncovered_geometry (bool): Suppress an error if some geometry parts aren't covered
            by any OSM extract. Defaults to `False`.

    Raises:
        RuntimeError: If provided extracts index is empty.
        RuntimeError: If there is no extracts covering a given geometry (singularly or in group).
        ValueError: If geometry_coverage_iou_threshold is outside bounds [0, 1].

    Returns:
        Set[str]: Selected extract index string values.
    """
    if polygons_index_gdf is None:
        raise RuntimeError("Extracts index is empty.")

    if geometry_coverage_iou_threshold < 0 or geometry_coverage_iou_threshold > 1:
        raise ValueError("geometry_coverage_iou_threshold is outside required bounds [0, 1]")

    checked_extracts_ids, iou_metric_values = _cover_geometry_with_extracts(
        geometry=geometry,
        polygons_index_gdf=polygons_index_gdf,
        allow_uncovered_geometry=allow_uncovered_geometry,
    )

    selected_extracts_ids: set[str] = set()
    for extract_id, iou_metric_value in zip(checked_extracts_ids, iou_metric_values):
        if iou_metric_value >= geometry_coverage_iou_threshold or not selected_extracts_ids:
            selected_extracts_ids.add(extract_id)
        else:
            skipped_extract = polygons_index_gdf[polygons_index_gdf.id == extract_id].iloc[0]
            warnings.warn(
                (
                    "Skipping extract because of low IoU value "
                    f"({skipped_extract.file_name}, {iou_metric_value:.3g})."
                ),
                GeometryNotCoveredWarning,
                stacklevel=0,
            )

    return selected_extracts_ids


def _cover_geometry_with_extracts(
    geometry: BaseGeometry,
    polygons_index_gdf: gpd.GeoDataFrame,
    allow_uncovered_geometry: bool = False,
) -> tuple[list[str], list[float]]:
    """
    Intersect a geometry with extracts and return the IoU coverage.

    Args:
        geometry (BaseGeometry): Geometry to be covered.
        polygons_index_gdf (gpd.GeoDataFrame): Index of available extracts.
        allow_uncovered_geometry (bool): Suppress an error if some geometry parts aren't covered
            by any OSM extract. Defaults to `False`.

    Raises:
        RuntimeError: If provided extracts index is empty.
        RuntimeError: If there is no extracts covering a given geometry (singularly or in group).

    Returns:
        tuple[list[str], list[float]]: List of extracts index string values with a list
            of IoU metric values.
    """
    if polygons_index_gdf is None:
        raise RuntimeError("Extracts index is empty.")

    checked_extracts_ids: list[str] = []
    iou_metric_values: list[float] = []

    if geometry.geom_type == "Polygon":
        geometry_to_cover = geometry.buffer(0)
    else:
        geometry_to_cover = geometry.buffer(1e-6)

    exactly_matching_geometry = polygons_index_gdf.loc[
        polygons_index_gdf.geometry.geom_equals_exact(geometry, tolerance=1e-6)
    ]
    if len(exactly_matching_geometry) == 1:
        checked_extracts_ids.append(exactly_matching_geometry.iloc[0].id)
        iou_metric_values.append(1.0)
        return checked_extracts_ids, iou_metric_values

    while not geometry_to_cover.is_empty:
        matching_rows = polygons_index_gdf.loc[
            (~polygons_index_gdf["id"].isin(checked_extracts_ids))
            & (polygons_index_gdf.intersects(geometry_to_cover))
        ]
        if not len(matching_rows):
            if not allow_uncovered_geometry:
                raise GeometryNotCoveredError(
                    "Couldn't find extracts covering given geometry."
                    " If it's expected behaviour, you can suppress this error by passing"
                    " the `allow_uncovered_geometry=True` argument"
                    " or add `--allow-uncovered-geometry` flag to the CLI command."
                )
            warnings.warn(
                "Couldn't find extracts covering given geometry.",
                GeometryNotCoveredWarning,
                stacklevel=0,
            )
            break

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            geometry_intersection_area = matching_rows.geometry.intersection(geometry_to_cover).area
            matching_rows["iou_metric"] = geometry_intersection_area / (
                matching_rows.geometry.area + geometry_to_cover.area - geometry_intersection_area
            )

        best_matching_extract = matching_rows.sort_values(
            by=["iou_metric", "area"], ascending=[False, True]
        ).iloc[0]
        geometry_to_cover = geometry_to_cover.difference(best_matching_extract.geometry)
        checked_extracts_ids.append(best_matching_extract.id)
        iou_metric_values.append(best_matching_extract.iou_metric)

    return checked_extracts_ids, iou_metric_values


def _filter_extracts(
    geometry: BaseGeometry,
    extracts_ids: Iterable[str],
    polygons_index_gdf: gpd.GeoDataFrame,
    num_of_multiprocessing_workers: int,
    multiprocessing_activation_threshold: int,
) -> list[OpenStreetMapExtract]:
    """
    Filter a set of extracts to include least overlaps in it.

    Args:
        geometry (Union[BaseGeometry, BaseMultipartGeometry]): Geometry to be covered.
        extracts_ids (Iterable[str]): Group of selected extracts indexes.
        polygons_index_gdf (gpd.GeoDataFrame): Index of available extracts.
        num_of_multiprocessing_workers (int): Number of workers used for multiprocessing.
            Similar to `n_jobs` parameter from `scikit-learn` library.
        multiprocessing_activation_threshold (int): Number of gometries required to start
            processing on multiple processes.

    Raises:
        RuntimeError: If provided extracts index is empty.

    Returns:
        List[OpenStreetMapExtract]: Filtered list of extracts.
    """
    if polygons_index_gdf is None:
        raise RuntimeError("Extracts index is empty.")

    sorted_extracts_gdf = polygons_index_gdf.loc[
        polygons_index_gdf["id"].isin(extracts_ids)
    ].sort_values(by="area", ignore_index=True, ascending=False)

    filtered_extracts: list[OpenStreetMapExtract] = []
    filtered_extracts_ids: set[str] = set()

    geometries = _flatten_geometry(geometry)

    total_geometries = len(geometries)

    if (
        num_of_multiprocessing_workers > 1
        and total_geometries >= multiprocessing_activation_threshold
    ):
        filter_extracts_func = partial(
            _filter_extracts_for_single_geometry,
            sorted_extracts_gdf=sorted_extracts_gdf,
        )

        for extract_ids_list in process_map(
            filter_extracts_func,
            geometries,
            desc="Filtering extracts",
            max_workers=num_of_multiprocessing_workers,
            chunksize=ceil(total_geometries / (4 * num_of_multiprocessing_workers)),
            disable=FORCE_TERMINAL,
        ):
            filtered_extracts_ids.update(extract_ids_list)
    else:
        for sub_geometry in geometries:
            filtered_extracts_ids.update(
                _filter_extracts_for_single_geometry(sub_geometry, sorted_extracts_gdf)
            )

    simplified_extracts_ids = _simplify_selected_extracts(
        filtered_extracts_ids, sorted_extracts_gdf
    )

    for extract_row in sorted_extracts_gdf.loc[
        sorted_extracts_gdf["id"].isin(simplified_extracts_ids)
    ].to_dict(orient="records"):
        extract = OpenStreetMapExtract(
            id=extract_row["id"],
            name=extract_row["name"],
            parent=extract_row["parent"],
            url=extract_row["url"],
            geometry=extract_row["geometry"],
            file_name=extract_row["file_name"],
        )
        filtered_extracts.append(extract)

    return filtered_extracts


def _filter_extracts_for_single_geometry(
    geometry: BaseGeometry, sorted_extracts_gdf: gpd.GeoDataFrame
) -> set[str]:
    """
    Filter a set of extracts to include least overlaps in it for a single geometry.

    Works by selecting biggest extracts (by area) and not including smaller ones if they don't
    increase a coverage.

    Args:
        geometry (BaseGeometry): Geometry to be covered.
        sorted_extracts_gdf (gpd.GeoDataFrame): Sorted index of available extracts.

    Returns:
        Set[str]: Selected extract index string values.
    """
    filtered_extracts_ids: set[str] = set()

    if geometry.geom_type == "Polygon":
        geometry_to_cover = geometry.buffer(0)
    else:
        geometry_to_cover = geometry.buffer(1e-6)

    for _, extract_row in sorted_extracts_gdf.iterrows():
        if geometry_to_cover.is_empty:
            break

        if extract_row.geometry.disjoint(geometry_to_cover):
            continue

        geometry_to_cover = geometry_to_cover.difference(extract_row.geometry)
        filtered_extracts_ids.add(extract_row.id)

    return filtered_extracts_ids


def _simplify_selected_extracts(
    filtered_extracts_ids: set[str], sorted_extracts_gdf: gpd.GeoDataFrame
) -> set[str]:
    simplified_extracts_ids: set[str] = filtered_extracts_ids.copy()

    matching_extracts = sorted_extracts_gdf.loc[
        sorted_extracts_gdf["id"].isin(simplified_extracts_ids)
    ]

    simplify_again = True
    while simplify_again:
        simplify_again = False
        extract_to_remove = None
        for extract_id in simplified_extracts_ids:
            extract_geometry = (
                matching_extracts.loc[sorted_extracts_gdf["id"] == extract_id].iloc[0].geometry
            )
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=FutureWarning)
                other_geometries_gdf = matching_extracts.loc[
                    sorted_extracts_gdf["id"] != extract_id
                ]
                if GEOPANDAS_NEW_API:
                    other_geometries = other_geometries_gdf.union_all()
                else:
                    other_geometries = other_geometries_gdf.unary_union
            if extract_geometry.covered_by(other_geometries):
                extract_to_remove = extract_id
                simplify_again = True
                break

        if extract_to_remove is not None:
            simplified_extracts_ids.remove(extract_to_remove)

    return simplified_extracts_ids


def _flatten_geometry(geometry: BaseGeometry) -> list[BaseGeometry]:
    """Flatten all geometries into a list of BaseGeometries."""
    if isinstance(geometry, BaseMultipartGeometry):
        geometries = []
        for sub_geom in geometry.geoms:
            geometries.extend(_flatten_geometry(sub_geom))
        return geometries
    return [geometry]


find_smallest_containing_extract = deprecate(
    "find_smallest_containing_extract",
    find_smallest_containing_extracts,
    "0.9.0",
    msg="Use `find_smallest_containing_extracts` instead. Deprecated since 0.9.0 version.",
)



import multiprocessing
import traceback
from pathlib import Path
from queue import Empty, Queue
from time import sleep
from typing import Any, Callable, Optional

import pyarrow as pa
import pyarrow.parquet as pq

from quackosm._exceptions import MultiprocessingRuntimeError
from quackosm._rich_progress import TaskProgressBar

# Using `spawn` method to enable integration with Polars and probably other Rust-based libraries
# https://docs.pola.rs/user-guide/misc/multiprocessing/
ctx: multiprocessing.context.SpawnContext = multiprocessing.get_context("spawn")


def _job(
    queue: Queue[tuple[str, int]],
    save_path: Path,
    function: Callable[[pa.Table], pa.Table],
    columns: Optional[list[str]] = None,
) -> None:  # pragma: no cover
    current_pid = multiprocessing.current_process().pid

    filepath = save_path / f"{current_pid}.parquet"
    writer = None
    while not queue.empty():
        try:
            file_name, row_group_index = None, None
            file_name, row_group_index = queue.get_nowait()

            pq_file = pq.ParquetFile(file_name)
            row_group_table = pq_file.read_row_group(row_group_index, columns=columns)
            if len(row_group_table) == 0:
                continue

            result_table = function(row_group_table)

            if not writer:
                writer = pq.ParquetWriter(filepath, result_table.schema)

            writer.write_table(result_table)
        except Empty:
            pass
        except Exception as ex:
            if file_name is not None and row_group_index is not None:
                queue.put((file_name, row_group_index))

            msg = (
                f"Error in worker (PID: {current_pid},"
                f" Parquet: {file_name}, Row group: {row_group_index})"
            )
            raise MultiprocessingRuntimeError(msg) from ex

    if writer:
        writer.close()


class WorkerProcess(ctx.Process):  # type: ignore[name-defined,misc]
    def __init__(self, *args: Any, **kwargs: Any):
        multiprocessing.Process.__init__(self, *args, **kwargs)
        self._pconn, self._cconn = multiprocessing.Pipe()
        self._exception: Optional[tuple[Exception, str]] = None

    def run(self) -> None:  # pragma: no cover
        try:
            multiprocessing.Process.run(self)
            self._cconn.send(None)
        except Exception as e:
            tb: str = traceback.format_exc()
            self._cconn.send((e, tb))

    @property
    def exception(self) -> Optional[tuple[Exception, str]]:
        if self._pconn.poll():
            self._exception = self._pconn.recv()
        return self._exception


def map_parquet_dataset(
    dataset_path: Path,
    destination_path: Path,
    function: Callable[[pa.Table], pa.Table],
    columns: Optional[list[str]] = None,
    progress_bar: Optional[TaskProgressBar] = None,
) -> None:
    """
    Apply a function over parquet dataset in a multiprocessing environment.

    Will save results in multiple files in a destination path.

    Args:
        dataset_path (Path): Path of the parquet dataset.
        destination_path (Path): Path of the destination.
        function (Callable[[pa.Table], pa.Table]): Function to apply over a row group table.
            Will save resulting table in a new parquet file.
        columns (Optional[list[str]]): List of columns to read. Defaults to `None`.
        progress_bar (Optional[TaskProgressBar]): Progress bar to show task status.
            Defaults to `None`.
    """
    dataset = pq.ParquetDataset(dataset_path)

    tuples_to_queue = []
    for pq_file in dataset.files:
        for row_group in range(pq.ParquetFile(pq_file).num_row_groups):
            tuples_to_queue.append((pq_file, row_group))

    total = len(tuples_to_queue)
    if progress_bar:  # pragma: no cover
        progress_bar.create_manual_bar(total=total)

    queue: Queue[tuple[str, int]] = ctx.Manager().Queue()
    for queue_tuple in tuples_to_queue:
        queue.put(queue_tuple)

    destination_path.mkdir(parents=True, exist_ok=True)

    try:
        processes = [
            WorkerProcess(
                target=_job,
                args=(queue, destination_path, function, columns),
            )
            for _ in range(min(multiprocessing.cpu_count(), total))
        ]
        _run_processes(processes=processes, queue=queue, total=total, progress_bar=progress_bar)
    finally:  # pragma: no cover
        _report_exceptions(processes=processes)


def _run_processes(
    processes: list[WorkerProcess],
    queue: Queue[tuple[str, int]],
    total: int,
    progress_bar: Optional[TaskProgressBar],
) -> None:
    # Run processes
    for p in processes:
        if queue.empty():
            break
        p.start()

    sleep_time = 0.1
    while any(process.is_alive() for process in processes):
        if any(p.exception for p in processes):  # pragma: no cover
            break

        if progress_bar:  # pragma: no cover
            progress_bar.update_manual_bar(current_progress=total - queue.qsize())

        sleep(sleep_time)
        sleep_time = min(1.0, sleep_time + 0.1)

    if progress_bar:  # pragma: no cover
        progress_bar.update_manual_bar(current_progress=total)


def _report_exceptions(processes: list[WorkerProcess]) -> None:
    # In case of exception
    exceptions = []
    for p in processes:
        if p.is_alive():
            p.terminate()

        if p.exception:
            exceptions.append(p.exception)

    if exceptions:
        # use ExceptionGroup in Python3.11
        _raise_multiple(exceptions)


def _raise_multiple(exceptions: list[tuple[Exception, str]]) -> None:
    if not exceptions:
        return
    try:
        error, traceback = exceptions.pop()
        msg = f"{error}\n\nOriginal {traceback}"
        raise type(error)(msg)
    finally:
        _raise_multiple(exceptions)



"""Module contains a dedicated type alias for OSM tags filter."""

from collections.abc import Iterable, Mapping
from typing import Union, cast, overload

from quackosm._typing import is_expected_type

OsmTagsFilter = Mapping[str, Union[list[str], str, bool]]

GroupedOsmTagsFilter = Mapping[str, OsmTagsFilter]


@overload
def merge_osm_tags_filter(osm_tags_filter: OsmTagsFilter) -> OsmTagsFilter: ...


@overload
def merge_osm_tags_filter(osm_tags_filter: GroupedOsmTagsFilter) -> OsmTagsFilter: ...


@overload
def merge_osm_tags_filter(osm_tags_filter: Iterable[OsmTagsFilter]) -> OsmTagsFilter: ...


@overload
def merge_osm_tags_filter(osm_tags_filter: Iterable[GroupedOsmTagsFilter]) -> OsmTagsFilter: ...


def merge_osm_tags_filter(
    osm_tags_filter: Union[
        OsmTagsFilter, GroupedOsmTagsFilter, Iterable[OsmTagsFilter], Iterable[GroupedOsmTagsFilter]
    ],
) -> OsmTagsFilter:
    """
    Merge OSM tags filter into `OsmTagsFilter` type.

    Optionally merges `GroupedOsmTagsFilter` into `OsmTagsFilter` to allow loaders to load all
    defined groups during single operation.

    Args:
        osm_tags_filter: OSM tags filter definition.

    Raises:
        AttributeError: When provided tags don't match both
            `OsmTagsFilter` or `GroupedOsmTagsFilter`.

    Returns:
        OsmTagsFilter: Merged filters.
    """
    if is_expected_type(osm_tags_filter, OsmTagsFilter):
        return cast(OsmTagsFilter, osm_tags_filter)
    elif is_expected_type(osm_tags_filter, GroupedOsmTagsFilter):
        return _merge_grouped_osm_tags_filter(cast(GroupedOsmTagsFilter, osm_tags_filter))
    elif is_expected_type(osm_tags_filter, Iterable):
        return _merge_multiple_osm_tags_filters(
            [
                merge_osm_tags_filter(
                    cast(Union[OsmTagsFilter, GroupedOsmTagsFilter], sub_osm_tags_filter)
                )
                for sub_osm_tags_filter in osm_tags_filter
            ]
        )

    raise AttributeError(
        "Provided tags don't match required type definitions"
        " (OsmTagsFilter or GroupedOsmTagsFilter)."
    )


def merge_key_value_pairs_to_osm_tags_filter(
    key_value_pairs: list[tuple[str, Union[list[str], str, bool]]],
) -> OsmTagsFilter:
    merged_osm_tag_filter = _merge_multiple_osm_tags_filters(
        {osm_tag_filter_key: osm_tag_filter_value}
        for osm_tag_filter_key, osm_tag_filter_value in key_value_pairs
    )
    return merged_osm_tag_filter


def check_if_any_osm_tags_filter_value_is_positive(
    osm_tags_filter: Union[GroupedOsmTagsFilter, OsmTagsFilter],
) -> bool:
    if is_expected_type(osm_tags_filter, OsmTagsFilter):
        return any(
            osm_tag_filter_value != False  # noqa: E712
            for osm_tag_filter_value in cast(OsmTagsFilter, osm_tags_filter).values()
        )
    elif is_expected_type(osm_tags_filter, GroupedOsmTagsFilter):
        return any(
            check_if_any_osm_tags_filter_value_is_positive(osm_tags_filter_group)
            for osm_tags_filter_group in cast(GroupedOsmTagsFilter, osm_tags_filter).values()
        )

    raise AttributeError(
        "Provided tags don't match required type definitions"
        " (OsmTagsFilter or GroupedOsmTagsFilter)."
    )


def _merge_grouped_osm_tags_filter(grouped_filter: GroupedOsmTagsFilter) -> OsmTagsFilter:
    """
    Merge grouped osm tags filter into a base one.

    Function merges all filter categories into a single one for an OSM loader to use.

    Args:
        grouped_filter (GroupedOsmTagsFilter): Grouped filter to be merged into a single one.

    Returns:
        osm_tags_type: Merged filter.
    """
    if not is_expected_type(grouped_filter, GroupedOsmTagsFilter):
        raise ValueError(
            "Provided filter doesn't match required `GroupedOsmTagsFilter` definition."
        )

    return _merge_multiple_osm_tags_filters(grouped_filter.values())


def _merge_multiple_osm_tags_filters(osm_tags_filters: Iterable[OsmTagsFilter]) -> OsmTagsFilter:
    """
    Merge multiple osm tags filters into a single one.

    Function merges all OsmTagsFilters into a single one for an OSM loader to use.

    Args:
        osm_tags_filters (Iterable[OsmTagsFilter]): List of filters to be merged into a single one.

    Returns:
        osm_tags_type: Merged filter.
    """
    if not is_expected_type(osm_tags_filters, Iterable[OsmTagsFilter]):
        raise ValueError(
            "Provided filter doesn't match required `Iterable[OsmTagsFilter]` definition."
        )

    result: dict[str, Union[list[str], str, bool]] = {}
    for osm_tags_filter in osm_tags_filters:
        for osm_tag_key, osm_tag_value in osm_tags_filter.items():
            if osm_tag_key not in result:
                result[osm_tag_key] = []

            is_current_value_positive = (
                isinstance(result[osm_tag_key], (list, bool)) and result[osm_tag_key]
            )
            is_current_value_negative = result[osm_tag_key] == False  # noqa: E712

            # If current value is negative and filter already have positive value
            # If current value is positive and filter already have negative value
            if (is_current_value_positive and osm_tag_value == False) or (  # noqa: E712
                is_current_value_negative and osm_tag_value != False  # noqa: E712
            ):
                raise ValueError(
                    "Provided OSM tags filter values cannot be merged.\n"
                    f"There is a conflict between the following values with '{osm_tag_key}' key:"
                    f" {result[osm_tag_key]} and {osm_tag_value}."
                )

            # If filter is already a positive boolean, skip
            if isinstance(result[osm_tag_key], bool) and result[osm_tag_key]:
                continue

            current_values_list = cast(list[str], result[osm_tag_key])

            # Check bool
            if osm_tag_value == True:  # noqa: E712
                result[osm_tag_key] = True
            elif osm_tag_value == False:  # noqa: E712
                result[osm_tag_key] = False
            # Check string
            elif isinstance(osm_tag_value, str) and osm_tag_value not in current_values_list:
                current_values_list.append(osm_tag_value)
            # Check list
            elif isinstance(osm_tag_value, list):
                new_values = [value for value in osm_tag_value if value not in current_values_list]
                current_values_list.extend(new_values)

    return cast(OsmTagsFilter, result)



"""
OpenStreetMap extracts tree.

This module contains function for displaying available extracts in the tree form.
"""

from typing import Callable

import pandas as pd
from rich.tree import Tree

from quackosm.osm_extracts.extract import OsmExtractSource


def get_available_extracts_as_rich_tree(
    source_enum: OsmExtractSource,
    osm_extract_source_index_functions: dict[OsmExtractSource, Callable[[], pd.DataFrame]],
    use_full_names: bool = False,
) -> Tree:
    """Transform available OSM extracts into a tree from the Rich library."""
    if source_enum == OsmExtractSource.any:
        root = Tree("All extracts")
        for other_source_enum, get_index_function in osm_extract_source_index_functions.items():
            branch_id = other_source_enum.value
            branch = root.add(branch_id)
            create_rich_tree_branch(branch_id, branch, get_index_function(), use_full_names)
    else:
        root_id = source_enum.value
        root = Tree(root_id)
        create_rich_tree_branch(
            root_id, root, osm_extract_source_index_functions[source_enum](), use_full_names
        )

    return root


def create_rich_tree_branch(
    parent_id: str, tree: Tree, index: pd.DataFrame, use_full_names: bool, depth: int = 0
) -> None:
    """
    Iterate OSM extracts recursively and create tree branches.

    Args:
        parent_id (str): Current id of the extract used to find children.
        tree (Tree): Tree object from Rich library.
        index (pd.DataFrame): List of available OSM extracts.
        use_full_names (bool): Whether to display full name, or short name of the extract.
            Full name contains all parents of the extract.
        depth (int, optional): Depth of the recursion. Defaults to 0.
    """
    matching_children = index[index.parent == parent_id].sort_values(by="name")
    for matching_child in matching_children.to_dict(orient="records"):
        name = matching_child["file_name"] if use_full_names else matching_child["name"]
        url = matching_child["url"]
        area = human_format(matching_child["area"])
        branch = tree.add(f":globe_with_meridians: [link={url}]{name}[/link] ({area} km\u00b2)")

        create_rich_tree_branch(matching_child["id"], branch, index, use_full_names, depth + 1)

    if depth == 0:
        loose_parents = sorted(set(index["parent"]).difference(index["id"]).difference([parent_id]))
        for loose_parent in loose_parents:
            branch = tree.add(loose_parent)
            create_rich_tree_branch(loose_parent, branch, index, use_full_names, depth + 1)


def human_format(num: float) -> str:
    """Change big number into a human readable format."""
    num = float(f"{num:.3g}")
    magnitude = 0
    while abs(num) >= 1000:
        magnitude += 1
        num /= 1000.0
    return "{}{}".format(
        f"{num:f}".rstrip(".0"),
        ["", "K", "M", "B"][magnitude],
    )



from functools import partial
from pathlib import Path
from typing import Optional

import geoarrow.pyarrow as ga
import pyarrow as pa
from shapely import STRtree
from shapely.geometry.base import BaseGeometry

from quackosm._parquet_multiprocessing import map_parquet_dataset
from quackosm._rich_progress import TaskProgressBar


def _intersect_nodes(
    table: pa.Table,
    geometry_filter: BaseGeometry,
) -> pa.Table:  # pragma: no cover
    points_array = ga.to_geopandas(
        ga.point().from_geobuffers(
            None,
            x=table["lon"].to_numpy(),
            y=table["lat"].to_numpy(),
        )
    )

    tree = STRtree(points_array)

    intersecting_ids_array = table["id"].take(tree.query(geometry_filter, predicate="intersects"))

    return pa.table({"id": intersecting_ids_array})


def intersect_nodes_with_geometry(
    tmp_dir_path: Path,
    geometry_filter: BaseGeometry,
    progress_bar: Optional[TaskProgressBar] = None,
) -> None:
    """
    Intersects nodes points with geometry filter using spatial index with multiprocessing.

    Args:
        tmp_dir_path (Path): Path of the working directory.
        geometry_filter (BaseGeometry): Geometry used for filtering.
        progress_bar (Optional[TaskProgressBar]): Progress bar to show task status.
            Defaults to `None`
    """
    dataset_path = tmp_dir_path / "nodes_valid_with_tags"
    destination_path = tmp_dir_path / "nodes_intersecting_ids"

    map_parquet_dataset(
        dataset_path=dataset_path,
        destination_path=destination_path,
        progress_bar=progress_bar,
        function=partial(_intersect_nodes, geometry_filter=geometry_filter),
        columns=["id", "lat", "lon"],
    )
